 

import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.text.*;
import javax.swing.undo.StateEditable;

/**
 * Classe Cell impl?mentant une cellule d'un tableur (SpreadSheet).
 * 
 * @author Denis Di Filippantonio for Dynasite
 * @date 13/04/2003
 * @version 1.1
 */
class Cell extends JTextPane implements CellInterf, StateEditable {
	//************************ Variables d'instance ******************************
	//position de la cellule dans le tableau
	private Point position;
	//taille (nombre de ligne, nombre de colonne)
	private Dimension size;
	// type de cellule
	private int type = TYPE_NUMBER;
	// formule initiale, avant calcul
	private String formula = "";
	// tableau contenant cette cellule
	private SpreadSheet spreadSheet;
	//nombre de sous-tableaux contenants cette cellule
	private int inSubTables = 0;
	// vecteur contenant les tableaux dont cette cellule est la premi?re case
	private ArrayList subTables = new ArrayList(1);
	// it?rateur sur le vecteur subTables, n?cessaire pour la conversion en HTML
	private Iterator tableIter;
	// hauteur minimum de la cellule pour que tout le texte soit visible
	private int minHeight = 0;

	//************************ Constructeur **************************************
	/**
	 * Constructeur g?n?ral.
	 *
	 * @param Point p: p.x >= 0, p.y >= 0 : position de la cellule
	 * @param Point s: s.x > 0, s.y > 0 : taille de la cellule
	 * @param MouseListener m: listener pour g?rer les actions sur les cellules
	 * @param UndoableEditListener u: pour le undo-redo des actions sur 
	 *																le texte de la cellule
	 * @param SpreadSheet sp: tableau contenant cette cellule
	 * @post une nouvelle cellule du tableau est cr?? et initialis?e correctement
	 */
	public Cell(
		Point p,
		Dimension s,
		MouseListener m,
		UndoableEditListener u,
		SpreadSheet sp) {
		position = new Point(p);
		size = new Dimension(s);
		addMouseListener(m);
		getDocument().addUndoableEditListener(u);
		getDocument().addDocumentListener(new TextChangedListener());
		spreadSheet = sp;
		setBorder(BorderFactory.createLineBorder(new Color(200, 200, 200)));
		setPreferredSize(new Dimension(100, 20));
	}

	/*****************************************************************************
	 * Constructeur identique au pr?c?dent mais pour une cellule de taille (1,1).
	 */
	public Cell(
		Point p,
		MouseListener m,
		UndoableEditListener u,
		SpreadSheet sp) {

		position = new Point(p);
		size = new Dimension(1, 1);
		addMouseListener(m);
		getDocument().addUndoableEditListener(u);
		getDocument().addDocumentListener(new TextChangedListener());
		spreadSheet = sp;
		setBorder(BorderFactory.createLineBorder(new Color(200, 200, 200)));
		setPreferredSize(new Dimension(100, 20));
	}

	/*****************************************************************************
	 * Constructeur cr?ant une cellule vide. Il s'agit en fait uniquement d'un
	 * pointeur vers une autre cellule (de position p) pour g?rer la fusion de 
	 * cellules.
	 */
	public Cell(Point p) {
		position = new Point(p);
		size = new Dimension(1, 1);
		setText("Empty Cell");
		type = TYPE_EMPTY;
	}

	//*************** M?thodes de l'interface CellInterf *************************
	public Point getLocation() {
		return new Point(position);
	}
	public void setLocation(Point p) {
		position = new Point(p);
	}
	public Dimension getCellSize() {
		return new Dimension(size);
	}
	public void setCellSize(Dimension s) {
		size = new Dimension(s);
	}
	public int getType() {
		return type;
	}

	public String toHTML(int width, int height) {
		if (type == TYPE_EMPTY)
			return "";

		String s = "<td";
		if (this.size.height != 1)
			s += " rowspan=" + this.size.height;
		if (this.size.width != 1)
			s += " colspan=" + this.size.width;
		s += " width=\"";
		s += (int) Math.round((100.0 * getSize().getWidth()) / width) + "%\"";
		s += " height=\"";
		s += (int) Math.round((100.0 * getSize().getHeight()) / height) + "%\"";

		StyledEditorKit kit = (StyledEditorKit) getEditorKit();
		AttributeSet attr = kit.getInputAttributes();
		if (StyleConstants.getAlignment(attr) == StyleConstants.ALIGN_CENTER)
			s += " align=center";
		if (StyleConstants.getAlignment(attr) == StyleConstants.ALIGN_RIGHT)
			s += " align=right";

		if (!getBackground().equals(Color.white))
			s += " bgcolor=" + getRGB(getBackground());
		s += ">";

		Font font = getFont();
		int size = (int) Math.round(font.getSize() / 6);
		s += "<font face=\"" + font.getName() + "\" size=\"" + size + "\"";
		if (!getForeground().equals(Color.black))
			s += " color=\"" + getRGB(getForeground()) + "\"";
		s += ">";

		boolean b = false;
		if (StyleConstants.isBold(attr)) {
			s += "<B>";
			b = true;
		}
		boolean i = false;
		if (StyleConstants.isItalic(attr)) {
			s += "<I>";
			i = true;
		}
		boolean u = false;
		if (StyleConstants.isUnderline(attr)) {
			s += "<U>";
			u = true;
		}

		s += getText().replaceAll("\n", "<br>");

		if (i)
			s += "</I>";
		if (b)
			s += "</B>";
		if (u)
			s += "</U>";

		s += "</font>";
		s += "</td>\n";
		return s;
	}

	public void fromHTML(String html) {
		int i = html.indexOf("<td");
		int j = html.indexOf(">", i);
		//*************** gestion de la couleur de fond *******
		String str = html.substring(i, j + 1);
		String color = getValue("bgcolor", str, false);
		if (!color.equals("")) {
			if (color.equals(getRGB(Color.green))) {
				setBackground(Color.green);
			} else if (color.equals(getRGB(Color.blue))) {
				setBackground(Color.blue);
			} else if (color.equals(getRGB(Color.yellow))) {
				setBackground(Color.yellow);
			} else if (color.equals(getRGB(Color.gray))) {
				setBackground(Color.gray);
			} else if (color.equals(getRGB(Color.red))) {
				setBackground(Color.red);
			} else if (color.equals(getRGB(Color.white))) {
				setBackground(Color.white);
			} else if (color.equals(getRGB(Color.black))) {
				setBackground(Color.black);
			} else
				System.out.println(
					"Cell ("
						+ position.x
						+ ","
						+ position.y
						+ "). Unknown backcolor: "
						+ color
						+ ".  Default backcolor selected: "
						+ getBackground());
		}
		//*************** gestion de l'alignement *************************
		String align = getValue("align", str, false);
		StyledEditorKit kit = (StyledEditorKit) getEditorKit();
		AttributeSet attr = kit.getInputAttributes();

		if (!align.equals("")) {
			if (align.equals("left"))
				setHorizontalAlignment(StyleConstants.ALIGN_LEFT);
			else if (align.equals("center"))
				setHorizontalAlignment(StyleConstants.ALIGN_CENTER);
			else if (align.equals("right"))
				setHorizontalAlignment(StyleConstants.ALIGN_RIGHT);
			else
				System.out.println(
					"Cell ("
						+ position.x
						+ ","
						+ position.y
						+ "). Unknown alignement: "
						+ align
						+ ".  Default alignment selected: "
						+ StyleConstants.getAlignment(attr));
		}
		//*************** gestion du style de la cellule ******************
		i = html.indexOf("<font", j);
		if (i != -1) {
			j = html.indexOf(">", i);
			str = html.substring(i, j + 1);
			//********* gestion de la police ****************
			// nouvelle m?thode
			String font = "";
			if (str.indexOf("\"") != -1)
				font = getValue("face", str, true);
			// ancienne m?thode
			else {
				int k = str.indexOf("face=");
				if (k != -1) {
					int o = str.indexOf("size");
					int m = str.indexOf("color");
					int n = str.indexOf(">");
					if (o == -1 && m != -1)
						n = m;
					else if (m == -1 && o != -1)
						n = o;
					else if (m != -1 && o != -1)
						n = Math.min(m, o);
					//trim: pour retirer les blancs au d?but et ? la fin
					font = str.substring(k + 5, n).trim();
				}
			}
			if (arrayContains(GraphicsEnvironment
				.getLocalGraphicsEnvironment()
				.getAvailableFontFamilyNames(),
				font)) {
				Font f = getFont();
				setFont(new Font(font, f.getStyle(), f.getSize()));
			} else
				System.out.println(
					"Cell ("
						+ position.x
						+ ","
						+ position.y
						+ "). Unknown font face: "
						+ font
						+ ".  Default font face selected: "
						+ getFont().getFontName());

			//********** gestion de la taille ***************
			String sSize = getValue("size", str, false);
			sSize = sSize.replaceAll("\"", "");

			if (!sSize.equals("")) {
				try {
					int size = Integer.parseInt(sSize);
					Font f = getFont();
					setFont(f.deriveFont((float) (6 * size)));
				} catch (NumberFormatException e) {
					System.out.println(
						"Cell ("
							+ position.x
							+ ","
							+ position.y
							+ "). Unreadable font size: "
							+ sSize
							+ ".  Default font size selected: "
							+ getFont().getSize());
				}
			}

			//********** gestion de la couleur ***************
			color = getValue("color", str, false);
			color = color.replaceAll("\"", "");

			if (!color.equals("")) {
				if (color.equals(getRGB(Color.green))) {
					setForeground(Color.green);
				} else if (color.equals(getRGB(Color.blue))) {
					setForeground(Color.blue);
				} else if (color.equals(getRGB(Color.yellow))) {
					setForeground(Color.yellow);
				} else if (color.equals(getRGB(Color.gray))) {
					setForeground(Color.gray);
				} else if (color.equals(getRGB(Color.red))) {
					setForeground(Color.red);
				} else if (color.equals(getRGB(Color.white))) {
					setForeground(Color.white);
				} else
					System.out.println(
						"Cell ("
							+ position.x
							+ ","
							+ position.y
							+ "). Unknown font color: "
							+ color
							+ ".  Default font color selected: "
							+ getForeground());
			}
		}
		//******** gestion du gras et italique ************
		if (html.indexOf("<B>") != -1) {
			setBold(true);
			j = html.indexOf(">", j + 1);
		}
		if (html.indexOf("<I>") != -1) {
			setItalic(true);
			j = html.indexOf(">", j + 1);
		}
		if (html.indexOf("<U>") != -1) {
			setUnderline(true);
			j = html.indexOf(">", j + 1);
		}
		//********** gestion du texte de la cellule *******
		int k = html.indexOf("</td>", j);
		if (html.indexOf("</B>") != -1) {
			k = Math.min(k, html.indexOf("</B>"));
		}
		if (html.indexOf("</I>") != -1) {
			k = Math.min(k, html.indexOf("</I>"));
		}
		if (html.indexOf("</font>") != -1) {
			k = Math.min(k, html.indexOf("</font>"));
		}
		String text = html.substring(j + 1, k);
		setText(((text.trim()).replaceAll("\n", "")).replaceAll("<br>", "\n"));
		updateHeight();

	}

	//**************** m?thodes de l'interface StateEditable *********************
	public void storeState(Hashtable state) {
		// gestion du style
		Font f = getFont();
		state.put("fontname", f.getFamily());
		state.put("fontsize", new Integer(f.getSize()));
		state.put("fontcolor", getForeground());
		state.put("cellcolor", getBackground());

		StyledEditorKit kit = (StyledEditorKit) getEditorKit();
		AttributeSet attr = kit.getInputAttributes();
		state.put("bold", new Boolean(StyleConstants.isBold(attr)));
		state.put("italic", new Boolean(StyleConstants.isItalic(attr)));
		state.put("underline", new Boolean(StyleConstants.isUnderline(attr)));
		state.put("align", new Integer(StyleConstants.getAlignment(attr)));

		// gestion des sous-tableaux
		state.put("insubtables", new Integer(inSubTables));
		state.put("subtables", subTables.clone());

		//gestion des fusions
		state.put("position", new Point(position));
		state.put("cellsize", new Dimension(size));
		state.put("pixsize", new Dimension(getPreferredSize()));
	}

	public void restoreState(Hashtable state) {
		Font f = getFont();

		String name =
			(state.containsKey("fontname")
				? (String) state.get("fontname")
				: f.getFamily());
		int fontSize =
			(state.containsKey("fontsize")
				? ((Integer) state.get("fontsize")).intValue()
				: f.getSize());

		Color fCol =
			(state.containsKey("fontcolor")
				? (Color) state.get("fontcolor")
				: getForeground());

		Color cCol =
			(state.containsKey("cellcolor")
				? (Color) state.get("cellcolor")
				: getBackground());

		SimpleAttributeSet sas = new SimpleAttributeSet();
		if (state.containsKey("bold")) {
			boolean under = ((Boolean) state.get("bold")).booleanValue();
			StyleConstants.setBold(sas, under);
		}
		if (state.containsKey("italic")) {
			boolean under = ((Boolean) state.get("italic")).booleanValue();
			StyleConstants.setItalic(sas, under);
		}
		if (state.containsKey("underline")) {
			boolean under = ((Boolean) state.get("underline")).booleanValue();
			StyleConstants.setUnderline(sas, under);
		}
		setCellCharAttributes(sas);
		if (state.containsKey("align")) {
			int align = ((Integer) state.get("align")).intValue();
			StyleConstants.setAlignment(sas, align);
		}
		setCellParAttributes(sas);

		setFont(new Font(name, Font.PLAIN, fontSize));
		setForeground(fCol);
		setBackground(cCol);

		inSubTables =
			(state.containsKey("insubtables")
				? ((Integer) state.get("insubtables")).intValue()
				: inSubTables);
		subTables =
			(state.containsKey("subtables")
				? (ArrayList) state.get("subtables")
				: subTables);
		// clone() n?cessaire (voir SpreadSheet.ArrayStateEdit.restoreState())

		position =
			(state.containsKey("position")
				? (new Point((Point) state.get("position")))
				: position);
		size =
			(state.containsKey("cellsize")
				? (new Dimension((Dimension) state.get("cellsize")))
				: size);
		setPreferredSize(
			(state.containsKey("pixsize")
				? (new Dimension((Dimension) state.get("pixsize")))
				: getPreferredSize()));
	}

	//********************* M?thodes publiques ***********************************

	/*****************************************************************************
	 * V?rifie si la cellule se trouve dans un sous-tableau.
	 *
	 * @return boolean: renvoit true si la cellule est dans un sous-tableau.
	 */
	public boolean isInSubTable() {
		return (inSubTables != 0);
	}

	/*****************************************************************************
	 * Renvoit le nombre de sous-tableaux dans lequel la cellule se trouve.
	 *
	 * @return int: le nombre de sous-tableaux.
	 */
	public int numSubTables() {
		return inSubTables;
	}

	/*****************************************************************************
	 * Informe la cellule qu'elle est plac?e dans un sous-tableau ca.
	 *
	 * @param CellArray ca: sous-tableau contenant cette cellule
	 * @post les variables inSubTables et subTables sont mises ? jour
	 * 			 pour tenir compte de l'ajout dans ca
	 */
	public void putInSubTable(CellArray ca) {
		inSubTables++;
		if (!(ca.getTLCell().equals(position)))
			return;
		// la cellule est la premi?re du sous-tableau
		// il faut ajouter ca au bon endroit pour que subTables soit ordonn?
		int i = 0;
		while (i < subTables.size()
			&& ((CellArray) subTables.get(i)).contains(ca)) {
			i++;
		}
		subTables.add(i, ca);
		//System.out.println(subTables.size());
	}

	/*****************************************************************************
	 * Informe la cellule qu'elle est retir?e du sous-tableau ca.
	 *
	 * @param CellArray ca: sous-tableau contenant cette cellule
	 * @post les variables inSubTables et subTables sont mises ? jour 
	 *		 	 pour tenir compte du retrait de ca
	 */
	public void removeFromSubTable(CellArray ca) {
		inSubTables--;
		if (subTables.contains(ca))
			subTables.remove(ca);
		//System.out.println(subTables.size());
	}

	/*****************************************************************************
	 * Initialisation de l'export. On doit recr?er l'iterator sur subTables avant
	 * d'effectuer l'export du tableau.
	 *
	 * @post tableIter a ?t? recr?? ? partir de subTables
	 */
	public void initExport() {
		tableIter = subTables.iterator();
	}

	/*****************************************************************************
	 * V?rifie si l'it?rateur des sous-tableaux dont cette cellule est la 
	 * premi?re case a encore des ?l?ments.
	 *
	 * @return boolean: true si il y a encore des sous-tableaux ? traiter
	 */
	public boolean hasNextSubTable() {
		return tableIter.hasNext();
	}

	/*****************************************************************************
	 * Renvoit le sous-tableau, dont cette cellule est la premi?re case,
	 * ? traiter suivant.
	 *
	 * @return CellArray: sous-tableau ? traiter suivant
	 */
	public CellArray nextSubTable() {
		return (CellArray) tableIter.next();
	}

	/*****************************************************************************
	 * Renvoit la hauteur minimum de la case permettant de voir tout 
	 * le contenu de celle-ci.
	 *
	 * @return int: la hauteur minimum.
	 */
	public int getMinHeight() {
		return minHeight;
	}

	/*****************************************************************************
	 * Met ? jour la hauteur de la ligne du tableau pour pouvoir voir tout 
	 * le contenu de cette celule. Si la hauteur est suffisante, aucune 
	 * modification n'est effectu?.
	 *
	 * @post - la ligne correspondante de spreadSheet est modifi?e en hauteur 
	 *		 - minHeight est mis ? jour
	 */
	public void updateHeight() {
		try {
			if ((!getText().equals("")) && getSize().height != 0) {
				Rectangle rect = modelToView(getDocument().getLength());
				minHeight = rect.y + rect.height + 3;
				if (getSize().height < minHeight)
					spreadSheet.setRowSize(
						position.x,
						minHeight - getSize().height);
			} else
				minHeight = 0;
		} catch (BadLocationException exc) {
			/*System.out.println("Automatic redimensionning Error.");
			System.out.println(
				"case("
					+ position
					+ ") minH:"
					+ minHeight
					+ " Size:"
					+ getSize().height);
			System.out.println(exc);*/
		}
	}

	private void setCellCharAttributes(AttributeSet attr) {
		StyledDocument doc = (StyledDocument) getDocument();
		if (doc == null)
			return;
		doc.setCharacterAttributes(0, doc.getLength(), attr, false);
		StyledEditorKit k = (StyledEditorKit) getEditorKit();
		MutableAttributeSet inputAttributes = k.getInputAttributes();
		inputAttributes.addAttributes(attr);
	}

	public void setBold() {
		StyledEditorKit kit = (StyledEditorKit) getEditorKit();
		MutableAttributeSet attr = kit.getInputAttributes();
		boolean b = (StyleConstants.isBold(attr)) ? false : true;
		setBold(b);
	}
	public void setBold(boolean b) {
		SimpleAttributeSet sas = new SimpleAttributeSet();
		StyleConstants.setBold(sas, b);
		setCellCharAttributes(sas);
	}

	public void setItalic() {
		StyledEditorKit kit = (StyledEditorKit) getEditorKit();
		MutableAttributeSet attr = kit.getInputAttributes();
		boolean b = (StyleConstants.isItalic(attr)) ? false : true;
		setItalic(b);
	}
	public void setItalic(boolean b) {
		SimpleAttributeSet sas = new SimpleAttributeSet();
		StyleConstants.setItalic(sas, b);
		setCellCharAttributes(sas);
	}

	public void setUnderline() {
		StyledEditorKit kit = (StyledEditorKit) getEditorKit();
		MutableAttributeSet attr = kit.getInputAttributes();
		boolean b = (StyleConstants.isUnderline(attr)) ? false : true;
		setUnderline(b);
	}
	public void setUnderline(boolean b) {
		SimpleAttributeSet sas = new SimpleAttributeSet();
		StyleConstants.setUnderline(sas, b);
		setCellCharAttributes(sas);
	}

	private void setCellParAttributes(AttributeSet attr) {
		StyledDocument doc = (StyledDocument) getDocument();
		if (doc == null)
			return;
		doc.setParagraphAttributes(0, doc.getLength(), attr, false);
		StyledEditorKit k = (StyledEditorKit) getEditorKit();
		MutableAttributeSet inputAttributes = k.getInputAttributes();
		inputAttributes.addAttributes(attr);
	}
	public void setHorizontalAlignment(int a) {
		SimpleAttributeSet sas = new SimpleAttributeSet();
		StyleConstants.setAlignment(sas, a);
		setCellParAttributes(sas);
	}

	//********************** M?thodes priv?es ************************************

	/*****************************************************************************
	 * Convertit un objet Color en code RGB ? utiliser en HTML
	 *
	 * @param Color col: la couleur ? convertir
	 * @return String: le code RGB
	 */
	private String getRGB(Color col) {
		String s;
		String c = Integer.toHexString(col.getRed());
		if (c.length() == 1)
			c = "0" + c;
		s = c;
		c = Integer.toHexString(col.getGreen());
		if (c.length() == 1)
			c = "0" + c;
		s += c;
		c = Integer.toHexString(col.getBlue());
		if (c.length() == 1)
			c = "0" + c;
		s += c;
		return s;
	}

	/*****************************************************************************
	 * Trouve la valeur de value dans le texte str, ce doit ?tre de la forme
	 * "value=qlqc". Renvoir le qlqc. Le qlqc se termine par un espace, le 
	 * symbole "plus petit" ("<") ou "plus grand" (">"). Le qlqc ne peut donc
	 * pas contenir un de ces symboles.
	 *
	 * @param String value: le texte dont il faut trouver la valeur
	 * @param String str: texte dans lequel il faut chercher
	 * @param boolean quotes : true si la valeur est entour?e par ""
	 * @return String: la valeur li?e ? value
	 */
	private String getValue(String value, String str, boolean quotes) {
		int i = str.indexOf(value);
		int k = str.indexOf("=", i);

		String val = "";
		if (k != -1 && i != -1) {
			if (quotes) {
				k++;
				val = str.substring(k + 1, str.indexOf("\"", k + 1));
			} else {
				StringTokenizer st =
					new StringTokenizer(str.substring(k + 1), " ><");
				val = st.nextToken();
			}
		}
		return val;
	}

	/*****************************************************************************
	 * V?rifie si un objet o se trouve dans un tableau d'objet arr.
	 *
	 * @param Object[] arr: tableau dans lequel il faut chercher
	 * @param Object o: objet ? chercher dans arr
	 * @return boolean: true si o se trouve dans arr
	 */
	private boolean arrayContains(Object[] arr, Object o) {
		for (int i = 0; i < arr.length; i++) {
			if (arr[i].equals(o))
				return true;
		}
		return false;
	}

	/*****************************************************************************
	 * Classe interne permettant de g?rer l'introduction de formules.
	 * A chaque modification du texte, elle v?rifie le contenu du texte
	 * pour voir si elle est du type TYPE_TEXT, TYPE_NUMBER ou TYPE_FORMULA.
	 * Elle ne s'occupe pas de calculer la valeur num?rique de la formule, qui
	 * doit ?tre effectu? uniquement quand on appui sur "Enter" ou qu'on 
	 * d?selectionne la cellule.
	 */
	private class TextChangedListener implements DocumentListener {
		public void insertUpdate(DocumentEvent e) {
			update();
		}
		public void removeUpdate(DocumentEvent e) {
			update();
		}
		public void changedUpdate(DocumentEvent e) {
			update();
		}

		private void update() {
			// redimensionne la case pour voir tout le texte (rajoute des lignes)
			updateHeight();

			/* ? r?activer quand n?cessaire (impl?mentation de formules)
			String text = getText();
			if (text.equals("")) 
				type = TYPE_NUMBER;
			try
			{
				Double.parseDouble(text);
				type = TYPE_NUMBER;
				return;
			}catch(NumberFormatException exc)
			{
				if (text.length() != 0 && text.charAt(0) != '=')
				{
					type = TYPE_TEXT;
					return;
				}
			}
			type = TYPE_FORMULA;*/
		}
	}
}
