 

import java.awt.*;
import java.util.Hashtable;
import javax.swing.undo.StateEditable;
import java.util.Hashtable;

/**
 * Classe CellArray impl?mentant la repr?sentation d'un ensemble de cellules
 * d'un sous-tableau (encadrement) d'un objet SpreadSheet.
 * 
 * @author Denis Di Filippantonio for Dynasite
 * @date 12/03/2003
 * @version 1.1
 */
class CellArray implements StateEditable
{
	
	//************************ Variables d'instance ******************************
	private Point tlCell;	//position de la premi?re cellule du sous-tableau
	private Point brCell;	//position de la derni?re cellule du sous-tableau
	
	//************************ Constructeur **************************************
	/**
	 * Constructeur
	 *
	 * @param Point tlC: tlC.x >= 0, tlC.y >= 0
	 * @param Point brC: brC.x >= 0, brC.y >= 0
	 * @post un nouvel objet CellArray est cr?? avec comme case sup?rieure gauche
	 * 			 tlC et case inf?rieure droite brC
	 */
	public CellArray(Point tlC, Point brC){
		tlCell = new Point(tlC);
		brCell = new Point(brC);
	}
	
	//*********************** M?thodes publiques *********************************
	/**
	 * Renvoit la position de la premi?re cellule du sous-tableau, le x est la
	 * position verticale et y est l'horizontale.
	 *
	 * @return Point: la position de la cellule
	 */
	public Point getTLCell(){						return new Point(tlCell);}
	
	/**
	 * Permet de changer la position de la premi?re cellule du sous-tableau.
	 *
	 * @param Point p: p.x >= 0, p.y >=0
	 * @post la premi?re cellule du sous-tableau est ? la position p
	 */
	public void setTLCell(Point p){				tlCell = new Point(p);}
	
	/**
	 * Renvoit la position de la derni?re cellule du sous-tableau, le x est 
	 * la position verticale et y est l'horizontale.
	 *
	 * @return Point: la position de la cellule
	 */
	public Point getBRCell(){						return new Point(brCell);}
	
	/**
	 * Permet de changer la position de la derni?re cellule du sous-tableau.
	 *
	 * @param Point p: p.x >= 0, p.y >=0
	 * @post la derni?re cellule du sous-tableau est ? la position p
	 */
	public void setBRCell(Point p){				brCell = new Point(p);}
	
	/**
	 * V?rifie si un autre sous-ensemble de cellules est compatible avec ce 
	 * sous-tableau. Deux sous-ensembles sont compatibles si leurs bords ne se
	 * croisent pas. Ils peuvent n?anmoins avoir des bords communs. Si le
	 * param?tre forFusion est true alors les deux sous-ensembles sont compar?s
	 * pour ensuite fusionner des cases qui peuvent se trouver ? l'int?rieure ou
	 * ? l'ext?rieure du sous-tableau. Dans ce cas, les sous-ensembles sont
	 * ?galement compatibles si ils sont ?gaux mais ne le sont pas si l'autre
	 * ensemble de cellules(celui qu'on fusionne) contient celui-ci(l'encadrement)
	 *
	 * @param Point tl: tl.x >=0, tl.y >=0 : premi?re case de l'autre sous-tableau
	 * @param Point br: br.x >=0, br.y >=0 : premi?re case de l'autre sous-tableau
	 * @param boolean forFusion: sp?cifie si c'est pour une fusion
	 * @return boolean: renvoit true si les deux sous-tableaux sont compatibles
	 */
	public boolean isCompatibleWith(Point tl, Point br, boolean forFusion)
	{
		// rectangle repr?sentant ce sous-tableau
		Rectangle rect = new Rectangle(tlCell, new Dimension(brCell.x-tlCell.x+1,
			brCell.y-tlCell.y+1));
		// rectangle repr?sentant l'autre sous-tableau
		Rectangle r = new Rectangle(tl, new Dimension(br.x-tl.x+1, br.y-tl.y+1));
		
		if (rect.equals(r) && (!forFusion))			return false;
		if (rect.equals(r) && forFusion)				return true;
		if (!rect.intersects(r))						return true;
		if (rect.contains(r))							return true;
		// r doit ?tre plus petit que rect si forFusion
		if (r.contains(rect) && (!forFusion))			return true;
		return false;
	}
	
	/**
	 * V?rifie si ce sous-tableau contient un autre sous-tableau ca, ceci incuant
	 * le cas o? les sous-tableaux sont identiques.
	 *
	 * @param CellArray ca: le sous-tableau avec lequel comparer
	 * @return boolean: renvoit true ce tableau contient ca
	 */
	public boolean contains(CellArray ca)
	{
		// rectangle repr?sentant ce sous-tableau
		Rectangle rect = new Rectangle(tlCell, new Dimension(brCell.x-tlCell.x+1,
			brCell.y-tlCell.y+1));
		// rectangle repr?sentant l'autre sous-tableau
		Rectangle r = new Rectangle(ca.getTLCell(), new Dimension(ca.getBRCell().x
			-ca.getTLCell().x+1, ca.getBRCell().y-ca.getTLCell().y+1));
		
		if (rect.contains(r))	return true;
		return false;
	}
	
	public void storeState(Hashtable state)
	{
		state.put("tlcell", getTLCell());
		state.put("brcell", getBRCell());
	}
	
	public void restoreState(Hashtable state)
	{
		if (state.containsKey("tlcell")) 
			setTLCell((Point)state.get("tlcell"));
		if (state.containsKey("brcell")) 
			setBRCell((Point)state.get("brcell"));
	}
}

