 

import java.awt.Point;
import java.awt.Dimension;

/**
 * Interface CellInterf d?finissant les m?thodes de base ex?cutables
 * sur les cellules (au sens g?n?ral) d'un tableau.
 * 
 * @author Denis Di Filippantonio for Dynasite
 * @date 13/03/2003
 * @version 1.1
 */
interface CellInterf{
	
	//******************** Variables statiques ***********************************
	// type de cellule
	static final int TYPE_EMPTY = 0; 			// cellule vide (pointeur)
	static final int TYPE_TEXT = 1; 			// cellule de type texte
	static final int TYPE_NUMBER = 2; 		// cellule de type num?rique
	static final int TYPE_FORMULA = 3; 		// cellule de type formule
	static final int TYPE_BORDER = 4; 		// bouton de bordure
	//****************************************************************************
	
	/**
	 * Renvoit la position de la cellule dans le tableau, le x est la position
	 * verticale et y est l'horizontale. Si la cellule est plus grande qu'une
	 * case (cellule fusionn?e) la position est celle de la case sup?rieure gauche
	 *
	 * @return Point: la position de la cellule
	 */
	public Point getLocation();
	
	/**
	 * Permet de changer la position d'une cellule (pour en fusionner plusieurs)
	 *
	 * @param Point p: p.x >= 0, p.y >=0
	 * @post la cellule est en position p
	 */
	public void setLocation(Point p);	
	
	/**
	 * Renvoit la taille de la cellule, en nombre de cases,
	 * la hauteur puis la largeur. Une cellule non-fusionn?e a une taille de (1,1)
	 *
	 * @return Dimension: la taille de la cellule
	 */
	public Dimension getCellSize();
	
	/**
	 * Permet de changer la taille d'une cellule (pour en fusionner plusieurs)
	 *
	 * @param Dimension p : p.x > 0 et p.y > 0
	 * @post la cellule a une taille p
	 */
	public void setCellSize(Dimension p);	

	/**
	 * Renvoit le type de cellule.
	 *
	 * @return int: le type de la cellule
	 */
	public int getType();
	
	/**
	 * Convertit le texte en HTML pass? en param?tre en propri?t? de la cellule
	 *
	 * @param String html: le texte html ? convertir, syntaxiquement correct
	 * @post la cellule a ?t? transform? conform?ment ? html
	 */
	public void fromHTML(String html);
	
	/**
	 * Effectue la conversion en HTML de la cellule
	 *
	 * @param int width > 0: largeur de la cellule en % du tableau
	 * @param int height > 0: hauteur de la cellule en % du tableau
	 * @return String: la repr?sentation en HTML de la cellule
	 */
	public String toHTML(int width, int height);
}
