 

import java.util.*;

/**
 * Classe HTMLTagParser permettant de "tokenizer" en string en fonction de tags 
 * HTML.
 * Par exemple, pour le code HTML:
 * <tr>
 *  <td> case 1 </td>
 * </tr>
 * <tr>
 *  <td> case 2 </td>
 * </tr>
 *
 * Tokeniz? selon le tag "td" donnera les ?l?ments:
 * <tr>
 *  <td> case 1 </td>
 * </tr>
 *
 * et
 * 
 * <tr>
 *  <td> case 2 </td>
 * </tr>
 * 
 * La s?paration se fait par "?tage", c'est-?-dire que si les tag de s?paration 
 * sont imbriqu?s ? plusieurs profondeur, on s?pare au m?me niveau:
 * <td>test1
 *  <td>test12</td>
 * </td>
 * <td>test2
 *  <td>test22</td>
 * </td>
 *
 * donnera (pour tag = td):
 * <td>test1
 *  <td>test12</td>
 * </td>
 * et
 * <td>test2
 *  <td>test22</td>
 * </td>
 * 
 * @author Denis Di Filippantonio
 * @date 15/12/2002
 * @version 1.0
 */

public class HTMLTagParser
{
	
	String text, tag;
	/**
	 * Constructeur: On cr?? un nouvel objet HTMLTagParser avec comme string ? traiter
	 * text et comme tag tag, tag ne doit pas contenir < et >. On suppose que les
	 * ?l?ments ? s?par?s sont enrour?es de: <tag et </tag> (<tag peut donc avoir des param?tres).
	 *
	 * @param: String text : le texte ? traiter
	 *				 String tag  : le tag d?limitant les ?l?ments
	 */
	public HTMLTagParser(String text, String tag)
	{
		this.text = new String(text);
		this.tag = tag;
	}
	
	public String next()
	{
		if (!hasNext())
			throw new NoSuchElementException();
		
		int lg = tag.length() + 3;
		int i = text.indexOf("<" + tag);
		int k = text.indexOf("<" + tag, i+lg);
		int j = text.indexOf("</" + tag + ">", i+lg);
		int last = j;
		int n = 1;
		
		while(j != -1 && n != 0)
		{
			while( k <= j && k != -1)
			{
				n++;
				k = text.indexOf("<" + tag, k+lg);
			}
			n--;
			last = j;
			j = text.indexOf("</" + tag + ">", j+lg);
		}
		if(n != 0)
			throw new NoSuchElementException("Error in the input String, number of" + 
				" opening and closing tags don't match.");
		
		String elem = text.substring(i, last+lg);
		// on peut ?liminer ce qu'il y a devant le i(si != 0) car on ne le renverra jamais
		text = text.substring(last+lg);
		return elem;
	}
	
	public boolean hasNext()
	{
		boolean b = text.indexOf("<" + tag) != -1;
		return (b && text.indexOf("</" + tag +">") != -1);
	}
	
	public void remove(){}
	
	/*public static void main(String[] args)
	{
		HTMLTagParser parser = new HTMLTagParser(args[0], args[1]);
		//System.out.println(args[0]);
		while (parser.hasNext())
		{
			System.out.println(parser.next() + "\nnext:\n");
		}
	}*/
}
