//Copyright 2004 Xavier Damman (xavier at tribal dot be)
//This file is part of phpDynaSite
//phpDynaSite is free software; you can redistribute it and/or modify
//it under the terms of the GNU General Public License as published by
//the Free Software Foundation; either version 2 of the License, or
//(at your option) any later version.
import java.util.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.undo.*;
import javax.swing.text.*;
import java.net.URL;
import java.net.MalformedURLException;

/**
 * Classe SpreadSheet, classe principale r?alisant un tableur en applet Java,
 * le tableur contient une barre d'outils et un ensemble de cellules.
 * Ce programme permet de g?n?rer du code HTML ? partir du tableau et de l'importer.
 * 
 * @author Denis Di Filippantonio for Dynasite
 * @date 14/04/2003
 * @version 1.1
 */
public class SpreadSheet extends JApplet {

	/* pour des tests:
	private ExternalWriter externalListener;
	public void addExternalListener(ExternalWriter ext) {
		externalListener = ext;
	}*/

	//************************ Variables statiques *******************************
	/** Hauteur par d?faut du tableau (sans la marge) */
	public static int defaultHeight = 5;
	/** Largeur par d?faut du tableau (sans la marge) */
	public static int defaultWidht = 3;
	/** URL ? partir de laquelle le programme peut ?tre ex?cut? (URL du programme 
	m?me, pas de la page HTML ! */
	private static final String[] allowedLocation = {"http://213.186.56.189/","http://localhost/","http://www.nivelles.be/","http://ns30818.ovh.net/"};

	//************************ Variables d'instance ******************************
	private JButton bBold,
		bItalic,
		bUnderline,
		bLeft,
		bCenter,
		bRight,
		bFusion,
		bUndo,
		bRedo,
		bBorder,
		bDebug;
	private JComboBox cbStyle, cbSize, cbFont, cbTextCol, cbBackCol;

	/* ensemble des cellules du tableur; les cellules fusionn?es apparaissent 
	 * plusieurs fois dans cette matrice (la vrai cellule et des cellules
	 * de TYPE_EMPTY; on ?crit d'abors l'indice de ligne, puis de colonne;
	 * elle ne contient pas les marges, les indices des cellules vont donc
	 * de 0 ? taille-1.
	 */
	private CellGrid cellGrid;
	/* Boutons des marges. SupButtons contient le bouton sup?rieur gauche
	 * en premi?re position. Par sym?trie, LeftButtons contient le m?me
	 * bouton en premi?re position.
	 */
	private ArrayList supButtons, leftButtons;
	// panneau contenant les cellules et les boutons de marges
	private JPanel cellPanel;
	// Listeners
	private KeyAction keyAction = new KeyAction();
	private MouseAction mouseAction = new MouseAction();
	private BorderAction borderAction;
	// liste des sous-tableaux (encadrements)
	private ArrayList subTableList = new ArrayList(1);
	// positions des cellules s?lectionn?es en haut ? gauche et en bas ? droite
	private Point tlSelectedCell, brSelectedCell;
	// premi?re cellule s?lectionn?e (celle sur laquelle on a cliqu?)
	private Cell firstCell;
	// gestionnaire du undo-redo
	private UndoHandler undoHandler;

	/**
	 * Gestion des styles pass?s en param?tres ? l'applet.
	 * Les styles sont form?s de :
	 *  0. Nom du style
	 *  1. Police
	 *  2. Taille
	 *  3. Couleur du texte
	 *  4. Alignement ('L' = gauche, 'C' = centr?, 'R' = droite)
	 *  5. Gras ('O' pour oui, 'N' pour non)
	 *  6. Italique ('O' pour oui, 'N' pour non)
	 *  7. Soulign? ('O' pour oui, 'N' pour non)
	 */
	private String[][] styles;
	private final String styleDelim = "|";
	private final String valueDelim = ",";
	private final String[] styleDefaut =
		{ "D?faut", "Verdana", "2", "Noir", "L", "N", "N", "N" };
	private String[] tableStyle = styleDefaut;

	//************************ M?thodes publiques ********************************

	/****************************************************************************
	 * M?thode d'initialisation de l'applet. Le tableur est initialis?, tous les
	 * ?l?ments sont cr??s et un tableau de taille (defaultHeight, defaultWidth)
	 * est cr??.
	 */
	private boolean bonneVersion = true;

	public void init() {
		if (System.getProperty("java.version").indexOf("1.4") == -1) {
			System.out.println(
				"*"
					+ System.getProperty("java.version"));
			bonneVersion = false;
			return;
		}
		bonneVersion = true;

/*		AllowedLocation thisScript= new AllowedLocation(getCodeBase().toString());
		if(!thisScript.isAllowed())
		{
			String errorMsg = thisScript.getErrorMsg();
			cellPanel = (JPanel) getContentPane();
			JTextPane textMsg = new JTextPane();
			textMsg.setText(errorMsg);
			cellPanel.add(textMsg);
			System.out.println(errorMsg);
			return;
		}*/

		// Chargement des styles pass?s en param?tres
		try {
			String stylesstring = getParameter("styles");
			Vector vvv = new Vector();
			StringTokenizer stok =
				new StringTokenizer(stylesstring, styleDelim, false);
			while (stok.hasMoreTokens()) {
				vvv.add(stok.nextToken());
			}
			styles = new String[vvv.size()][8];
			for (int i = 0; i < styles.length; i++) {
				stok =
					new StringTokenizer(
						(String) vvv.elementAt(i),
						valueDelim,
						false);
				int j = 0;
				try {
					while (stok.hasMoreTokens()) {
						styles[i][j++] = stok.nextToken();
					}
				} catch (Exception e) {
					System.out.println(
						"Erreur dans les styles pass?s en param?tres.");
					styles[i] = styleDefaut;
				}
			}
			tableStyle = styles[0];
		} catch (Exception e) {
			System.out.println(
				"Erreur dans les styles pass?s en param?tres ou "
					+ "non-pr?sents.");
			styles = new String[1][8];
			styles[0] = styleDefaut;
		}

		//******************************************
		setBackground(Color.white);

		getContentPane().setLayout(new BorderLayout());
		// barre d'outils
		getContentPane().add(createToolBar(), "North");
		// tableau
		FlowLayout f = new FlowLayout();
		f.setAlignment(FlowLayout.LEFT);
		JPanel p = new JPanel(f);
		p.setBackground(Color.white);
		cellPanel = new CellPanel();
		cellPanel.setBackground(Color.white);
		p.add(cellPanel);
		JScrollPane sp = new JScrollPane(p);
		getContentPane().add(sp, "Center");

		// Chargement du param?tre data
		// Xavier : 30/01/03
		String ImportData = getParameter("data");
		//try{
			if (!importFromHTML(ImportData))
				newSpreadSheet(defaultHeight, defaultWidht); //si l'import ?choue
		//}catch(Exception e){
		//	newSpreadSheet(defaultHeight, defaultWidht);
		//}
		
		addKeyListener(keyAction);
		cellPanel.addKeyListener(keyAction);

		setVisible(true);
	}

	/*****************************************************************************
	 * Cr?ation d'un nouveau tableau. Ce qu'il contient d?j? est effac?.
	 *
	 * @pre Le programme doit ?tre correctement initialis?: la barre d'outils
	 *			et le panneau d'affichage doivent ?tre pr?sents
	 * @param nbRows, nbCols : nombre de ligne, nombre de colonnes
	 */
	public void newSpreadSheet(int nbRows, int nbCols) {
		createCellArray(nbRows, nbCols);
		redrawCellPanel();
		// mise ? jour du style des cellules
		cbStyle.setSelectedIndex(0);

		for (int i = 0; i < getTabSize().height; i++) {
			for (int j = 0; j < getTabSize().width; j++) {
				setCellStyle(getCell(i, j));
			}
		}
		updateToolBar();
		undoHandler.discardAllEdits();
		undoHandler.setActive(true);
	}

	/*****************************************************************************
	 * Convertion du tableau en format HTML (sous forme de table).
	 *
	 * @pre Le programme doit ?tre correctement initialis?:  le panneau 
	 * 			d'affichage doit ?tre pr?sent et contenir des cellules
	 * @return String: la repr?sentation HTML de ce tableau
	 */
	public String exportToHTML() {
		// initialisation de l'export de chaque cellule
		for (int i = 0; i < getTabSize().height; i++) {
			for (int j = 0; j < getTabSize().width; j++) {
				getCell(i, j).initExport();
			}
		}

		String s =
			"<!--Created by SpreadSheet sizex="
				+ getTabSize().height
				+ " sizey="
				+ getTabSize().width
				+ " -->\n";

		// largeur et hauteur du tableau en pixel
		int w = (int) (cellPanel.getSize().getWidth());
		int h = (int) (cellPanel.getSize().getHeight());
		s += "<table width=" + w + " height=" + h;
		s += " CELLPADDING=\"0\" CELLSPACING=\"0\"" + ">\n";

		for (int i = 0; i < getTabSize().height; i++) {
			s += "<tr valign=top>\n";
			for (int j = 0; j < getTabSize().width; j++) {
				Cell c = getCell(i, j);
				if (c.hasNextSubTable())
					s += exportSubTable(c.nextSubTable(), w, h, 0);
				else if (!c.isInSubTable())
					s += c.toHTML(w, h);
			}
			s += "</tr>\n";
		}
		s += "</table>\n";
		return s;
	}

	/*****************************************************************************
	 * Importation d'un tableau ? partir d'un string au format HTML
	 * (sous forme de table). Ce code HTML doit ABSOLUMENT avoir ?t? g?n?r? par
	 * le programme lui-m?me !! Sinon, l'import sera annul?.
	 *
	 * @pre Le programme doit ?tre correctement initialis?:  le panneau 
	 * 			d'affichage doit ?tre pr?sent et contenir des cellules
	 * @post ce tableau est la repr?sentation de html, tous les ?l?ments
	 *			 sont initialis?s et l'affichage est mis ? jour
	 * @return boolean : true si la convertion a r?ussis
	 */
	public boolean importFromHTML(String html) {
		//********** initialisation du tableau ***************
		if (html == null)
			return false;
		
		int i = html.indexOf("<table");
		if (i == -1) {
			System.out.println(
				"Error of import: no table tag found. Import aborted.");
			return false;
		}

		int j = html.indexOf("<tr", i);
		if (j == -1) {
			System.out.println(
				"Error of import: no tr tag found. Import aborted.");
			return false;
		}

		String str = html.substring(0, j);
		// taille en pixel du tableau
		String widthS = getValue("width", str);
		String heightS = getValue("height", str);
		int width = 0;
		int height = 0;
		if (!widthS.equals("")) {
			width = (int) Float.parseFloat(widthS);
		}
		if (!heightS.equals("")) {
			height = (int) Float.parseFloat(heightS);
		}
		// taille en cellules du tableau
		String sizeXS = getValue("sizex", str);
		String sizeYS = getValue("sizey", str);
		if (sizeXS.equals("") || sizeYS.equals("")) {
			System.out.println(
				"Error of import: no table size indication found. Import aborted.");
			return false;
		}
		//(int)Float pour des raisons de compatibilit? ant?rieur
		int tableSizeX = (int) Float.parseFloat(sizeXS);
		int tableSizeY = (int) Float.parseFloat(sizeYS);

		//************* cr?ation du tableau ******************
		createCellArray(tableSizeX, tableSizeY);
		// initialise le style au style par d?faut s'il n'est pas red?finis apr?s
		cbStyle.setSelectedIndex(0);
		for (int k = 0; k < getTabSize().height; k++)
			for (int l = 0; l < getTabSize().width; l++)
				setCellStyle(getCell(k, l));

		/*
		 * le tableau est cr?? avec des cellules vides non-fusionn?es
		 * il faut maintenant le remplir avec les donn?es correctes
		 */
		HTMLTagParser pTR = new HTMLTagParser(html.substring(j), "tr");
		
		int rowNbr = 0;
		while (pTR.hasNext()) {
			String line = pTR.next();
			HTMLTagParser pTD = new HTMLTagParser(line, "td");
			int colNbr = 0;

			while (pTD.hasNext()) {
				String cellLine = pTD.next();
				Cell c = getCell(rowNbr, colNbr);
				// si une des condition est vrai c'est que cette case a d?j? ?t? trait?e
				// auparavant et il ne faut plus s'en occuper
				while (c.getType() == CellInterf.TYPE_EMPTY
					|| c.isInSubTable()) {
					colNbr++;
					c = getCell(rowNbr, colNbr);
				}
				importCell(cellLine, rowNbr, colNbr, width, height, 0);

				colNbr++;
			}
			rowNbr++;
		}

		//******** mise ? jour de l'affichage ******************
		redrawCellPanel();
		firstCell = getCell(0, 0);
		tlSelectedCell = firstCell.getLocation();
		brSelectedCell = firstCell.getLocation();
		repaint();
		undoHandler.discardAllEdits();
		updateToolBar();
		return true;
	}

	private void importCell(
		String cellLine,
		int rowNbr,
		int colNbr,
		int width,
		int height,
		int deepness) {
		String rowS = getValue("rowspan", cellLine);
		int row = (rowS.equals("") ? 1 : Integer.parseInt(rowS));
		String colS = getValue("colspan", cellLine);
		int col = (colS.equals("") ? 1 : Integer.parseInt(colS));

		//System.out.println("cell("+rowNbr+","+colNbr+"): "+cellLine);

		Cell cell = getCell(rowNbr, colNbr);

		if ((row != 1 || col != 1) && (cellLine.indexOf("<table") == -1)) {
			fusionCells(
				new Point(rowNbr, colNbr),
				new Point(rowNbr + row - 1, colNbr + col - 1));

			cell.setCellSize(new Dimension(col, row));
			setCell(cell, rowNbr, colNbr);
		}

		if (cellLine.indexOf("<table") != -1) {
			importSubTable(
				cellLine,
				rowNbr,
				colNbr,
				row,
				col,
				width,
				height,
				deepness);
		} else {
			cell.fromHTML(cellLine);

			String s = getValue("width", cellLine);
			if (width != 0 && !s.equals("")) {
				int i = s.indexOf("%\"");
				int cellWidth = Integer.parseInt(s.substring(1, i));
				cell.setPreferredSize(
					new Dimension(
						(cellWidth * width) / 100,
						cell.getPreferredSize().height));
			}

			s = getValue("height", cellLine);
			if (height != 0 && !s.equals("")) {
				int i = s.indexOf("%\"");
				int cellHeight = Integer.parseInt(s.substring(1, i));
				cell.setPreferredSize(
					new Dimension(
						cell.getPreferredSize().width,
						(cellHeight * height) / 100));
			}
		}
	}
	/*****************************************************************************
	
	*/
	private void importSubTable(
		String tableLine,
		int rowNbr,
		int colNbr,
		int row,
		int col,
		int width,
		int height,
		int deepness) {
		//********** initialisation du sous-tableau ***************

		Point tl = new Point(rowNbr, colNbr);
		Point br = new Point(rowNbr + row - 1, colNbr + col - 1);
		CellArray ca = new CellArray(tl, br);
		subTableList.add(ca);
		for (int i = tl.x; i <= br.x; i++) {
			for (int j = tl.y; j <= br.y; j++) {
				getCell(i, j).putInSubTable(ca);
			}
		}

		//permet de retirer un tag table (car il y en a deux imbriqu?s)
		HTMLTagParser pTable = new HTMLTagParser(tableLine, "tr");
		tableLine = pTable.next();
		pTable = new HTMLTagParser(tableLine, "table");
		tableLine = pTable.next();

		deepness++;

		//System.out.println("start subtable, deepness:"+deepness);
		/*
		 * le sous-tableau est cr??
		 * il faut maintenant le "remplir" avec les cellules
		 */
		HTMLTagParser pTR = new HTMLTagParser(tableLine, "tr");
		while (pTR.hasNext()) {
			String line = pTR.next();
			//System.out.println("rowNbr:"+rowNbr+"\n"+line);

			HTMLTagParser pTD = new HTMLTagParser(line, "td");
			int colN = colNbr;

			while (pTD.hasNext()) {
				String cellLine = pTD.next();
				Cell c = getCell(rowNbr, colN);
				//System.out.println("rowNbr:"+rowNbr+"colN:"+colN);
				while (c.getType() == CellInterf.TYPE_EMPTY
					|| c.numSubTables() > deepness) {
					colN++;
					c = getCell(rowNbr, colN);
				}
				//System.out.println("after iteration:"+"rowNbr:"+rowNbr+"colN:"+colN);
				importCell(cellLine, rowNbr, colN, width, height, deepness);

				colN++;
			}
			rowNbr++;
		}
		//System.out.println("end subtable, deepness:"+deepness);
	}

	/*****************************************************************************
	 * Renvoit la cellule correspondant ? la coordon?e p. Les positions des 
	 * cellules sont num?rot?es ? partir de 0
	 *
	 * @param Point p: p.x >= 0, p.y >= 0
	 * @return Cell: la cellule ? la position p
	 */
	Cell getCell(Point p) {
		return getCell(p.x, p.y);
	}

	/*****************************************************************************
	 * Renvoit la cellule correspondant ? la (i,j). Les positions des 
	 * cellules sont num?rot?es ? partir de 0
	 *
	 * @param int i, int j: i >= 0, j >= 0
	 * @return Cell: la cellule ? la position (i,j)
	 */
	Cell getCell(int i, int j) {
		return cellGrid.getCell(i, j);
	}

	//************************ M?thodes priv?es **********************************
	/*****************************************************************************
	 * Attribue la cellule c ? la position (i,j) du tableau. 
	 *
	 * @param int i, int j: i >= 0, j >= 0 et contenus dans le tableau
	 * @post la cellule c est ? la position (i,j)
	 */
	private void setCell(Cell c, int i, int j) {
		cellGrid.setCell(i, j, c);
	}

	/*****************************************************************************
	 * Renvoit la taille du tableur
	 *
	 * @return Dimension: une copie de la taille du tableau
	 */
	Dimension getTabSize() {
		return cellGrid.getSize();
	}

	/*****************************************************************************
	 * Cr?ation de la barre d'outils.
	 *
	 * @pre styles[][] doit ?tre initialis?
	 * @post tous les ?l?ments de la barre d'outils sont cr??s et initialis?s,
	 *			 avec leur ActionListener.
	 * @return JToolBar: la barre d'outils
	 */
	private JToolBar createToolBar() {
		JToolBar toolbar = new JToolBar();
		ToolBarAction tba = new ToolBarAction();
		toolbar.setBackground(Color.white);

		cbStyle = new JComboBox();
		for (int i = 0; i < styles.length; i++) {
			cbStyle.addItem(styles[i][0]);
		}
		cbStyle.setPreferredSize(new Dimension(45, 20));
		cbStyle.setRequestFocusEnabled(false);
		cbStyle.setFont(new Font(cbStyle.getFont().getName(), Font.PLAIN, 9));
		cbStyle.setSelectedIndex(0);
		toolbar.add(cbStyle);
		cbStyle.addItemListener(tba);

		String[] fonts =
			GraphicsEnvironment
				.getLocalGraphicsEnvironment()
				.getAvailableFontFamilyNames();
		cbFont = new JComboBox();
		for (int i = 0; i < fonts.length; i++) {
			cbFont.addItem(fonts[i]);
		}
		cbFont.setPreferredSize(new Dimension(90, 20));
		cbFont.setSelectedItem(tableStyle[1]);
		cbFont.setRequestFocusEnabled(false);
		cbFont.setFont(new Font(cbFont.getFont().getName(), Font.PLAIN, 9));
		toolbar.add(cbFont);
		cbFont.addItemListener(tba);

		cbSize = new JComboBox();
		cbSize.addItem("2");
		cbSize.addItem("3");
		cbSize.addItem("4");
		cbSize.addItem("5");
		cbSize.addItem("6");
		cbSize.addItem("7");
		cbSize.setPreferredSize(new Dimension(35, 20));
		cbSize.setFont(new Font(cbSize.getFont().getName(), Font.PLAIN, 9));
		cbSize.setEditable(false);
		cbSize.setRequestFocusEnabled(false);
		cbSize.setSelectedItem(tableStyle[2]);
		toolbar.add(cbSize);
		cbSize.addItemListener(tba);

		//toolbar.addSeparator();

		bBold =
			new JButton(
				new ImageIcon(
					getImage(getCodeBase(), "bold.gif"))) {
			public float getAlignmentY() {
				return 0.5f;
			}
		};
		bBold.setRequestFocusEnabled(false);
		bBold.setMargin(new Insets(1, 1, 1, 1));
		toolbar.add(bBold);
		bBold.addActionListener(tba);

		bItalic =
			new JButton(
				new ImageIcon(
					getImage(getCodeBase(), "italic.gif"))) {
			public float getAlignmentY() {
				return 0.5f;
			}
		};
		bItalic.setRequestFocusEnabled(false);
		bItalic.setMargin(new Insets(1, 1, 1, 1));
		toolbar.add(bItalic);
		bItalic.addActionListener(tba);

		bUnderline =
			new JButton(
				new ImageIcon(
					getImage(
						getCodeBase(),
						"underline.gif"))) {
			public float getAlignmentY() {
				return 0.5f;
			}
		};
		bUnderline.setRequestFocusEnabled(false);
		bUnderline.setMargin(new Insets(1, 1, 1, 1));
		toolbar.add(bUnderline);
		bUnderline.addActionListener(tba);

		//toolbar.addSeparator();

		bLeft =
			new JButton(
				new ImageIcon(
					getImage(getCodeBase(), "left.gif"))) {
			public float getAlignmentY() {
				return 0.5f;
			}
		};
		bLeft.setRequestFocusEnabled(false);
		bLeft.setMargin(new Insets(1, 1, 1, 1));
		toolbar.add(bLeft);
		bLeft.addActionListener(tba);

		bCenter =
			new JButton(
				new ImageIcon(
					getImage(getCodeBase(), "center.gif"))) {
			public float getAlignmentY() {
				return 0.5f;
			}
		};
		bCenter.setRequestFocusEnabled(false);
		bCenter.setMargin(new Insets(1, 1, 1, 1));
		toolbar.add(bCenter);
		bCenter.addActionListener(tba);

		bRight =
			new JButton(
				new ImageIcon(
					getImage(getCodeBase(), "right.gif"))) {
			public float getAlignmentY() {
				return 0.5f;
			}
		};
		bRight.setRequestFocusEnabled(false);
		bRight.setMargin(new Insets(1, 1, 1, 1));
		toolbar.add(bRight);
		bRight.addActionListener(tba);

		// toolbar.addSeparator();

		bFusion =
			new JButton(
				new ImageIcon(
					getImage(getCodeBase(), "fusion.gif"))) {
			public float getAlignmentY() {
				return 0.5f;
			}
		};
		bFusion.setRequestFocusEnabled(false);
		bFusion.setMargin(new Insets(1, 1, 1, 1));
		toolbar.add(bFusion);
		bFusion.addActionListener(tba);

		bBorder =
			new JButton(
				new ImageIcon(
					getImage(getCodeBase(), "border.gif"))) {
			public float getAlignmentY() {
				return 0.5f;
			}
		};
		bBorder.setRequestFocusEnabled(false);
		bBorder.setMargin(new Insets(1, 1, 1, 1));
		toolbar.add(bBorder);
		bBorder.addActionListener(tba);

		//toolbar.addSeparator();

		cbTextCol = new JComboBox();
		cbTextCol.addItem(Color.black);
		cbTextCol.addItem(Color.red);
		cbTextCol.addItem(Color.green);
		cbTextCol.addItem(Color.blue);
		cbTextCol.addItem(Color.yellow);
		cbTextCol.addItem(Color.gray);
		cbTextCol.addItem(Color.white);
		cbTextCol.setPreferredSize(new Dimension(38, 20));
		Color col = Color.black;
		if (tableStyle[3].equals("Vert")) {
			col = Color.green;
		} else if (tableStyle[3].equals("Bleu")) {
			col = Color.blue;
		} else if (tableStyle[3].equals("Jaune")) {
			col = Color.yellow;
		} else if (tableStyle[3].equals("Gris")) {
			col = Color.gray;
		} else if (tableStyle[3].equals("Rouge")) {
			col = Color.red;
		} else if (tableStyle[3].equals("Blanc")) {
			col = Color.white;
		}
		cbTextCol.setSelectedItem(col);
		cbTextCol.setRequestFocusEnabled(false);
		cbTextCol.setRenderer(new ListCellRenderer() {
			public Component getListCellRendererComponent(
				JList list,
				Object value,
				int index,
				boolean isSelected,
				boolean cellHasFocus) {
				return new ColoredText((Color) value);
			}
		});
		toolbar.add(cbTextCol);
		cbTextCol.addItemListener(tba);

		cbBackCol = new JComboBox();
		cbBackCol.addItem(Color.white);
		cbBackCol.addItem(Color.black);
		cbBackCol.addItem(Color.red);
		cbBackCol.addItem(Color.green);
		cbBackCol.addItem(Color.blue);
		cbBackCol.addItem(Color.yellow);
		cbBackCol.addItem(Color.gray);
		cbBackCol.setPreferredSize(new Dimension(38, 20));
		cbBackCol.setRequestFocusEnabled(false);
		cbBackCol.setSelectedIndex(0);
		cbBackCol.setRenderer(new ListCellRenderer() {
			public Component getListCellRendererComponent(
				JList list,
				Object value,
				int index,
				boolean isSelected,
				boolean cellHasFocus) {
				return new ColoredCell((Color) value);
			}
		});
		toolbar.add(cbBackCol);
		cbBackCol.addItemListener(tba);

		//toolbar.addSeparator();

		bUndo =
			new JButton(
				new ImageIcon(
					getImage(getCodeBase(), "undo.gif"))) {
			public float getAlignmentY() {
				return 0.5f;
			}
		};
		bUndo.setRequestFocusEnabled(false);
		bUndo.setMargin(new Insets(1, 1, 1, 1));
		bUndo.setEnabled(false);
		bUndo.addActionListener(tba);
		toolbar.add(bUndo);

		bRedo =
			new JButton(
				new ImageIcon(
					getImage(getCodeBase(), "redo.gif"))) {
			public float getAlignmentY() {
				return 0.5f;
			}
		};
		bRedo.setRequestFocusEnabled(false);
		bRedo.setMargin(new Insets(1, 1, 1, 1));
		bRedo.setEnabled(false);
		bRedo.addActionListener(tba);
		toolbar.add(bRedo);

		//toolbar.addSeparator();

		/*bDebug = new JButton("debug") {
			public float getAlignmentY() {
				return 0.5f;
			}
		};
		bDebug.setRequestFocusEnabled(false);
		bDebug.setMargin(new Insets(1, 1, 1, 1));
		toolbar.add(bDebug);
		bDebug.addActionListener(tba);*/

		//toolbar.add(Box.createHorizontalGlue());
		toolbar.setFloatable(false);
		return toolbar;

	}

	/*****************************************************************************
	 * Cr?ation de l'ensemble des cellules et disposition dans la matrice
	 * cellGrid. Le cellPanel n'est pas mis ? jour.
	 *
	 * @pre - int sizeX hauteur du tableau ? cr?er (sans marge)
	 *      - int sizeY largeur du tableau ? cr?er (sans marge)
	 * @post - cellGrid est rempli de nouvelles cellules de taille (1,1)
	 *			 - nouveau undoHandler et subTableList
	 *			 - la s?lection est initialis?e en (0,0)
	 */
	private void createCellArray(int sizeX, int sizeY) {
		cellGrid = new CellGrid(sizeX, sizeY);
		undoHandler = new UndoHandler();
		undoHandler.setActive(false);
		subTableList = new ArrayList(1);
		borderAction = new BorderAction();
		supButtons = new ArrayList(sizeY + 1);
		leftButtons = new ArrayList(sizeX + 1);

		// coin en haut ? gauche
		BorderButton b = new BorderButton(new Point(0, -1));
		b.addActionListener(borderAction);
		supButtons.add(b);
		leftButtons.add(b);

		//ligne de bordure gauche
		for (int i = 0; i < sizeX; i++) {
			b = new BorderButton(new Point(0, i));
			b.addActionListener(borderAction);
			b.addMouseMotionListener(borderAction);
			b.addMouseListener(borderAction);
			leftButtons.add(b);
		}

		//ligne de bordure sup?rieure
		for (int j = 0; j < sizeY; j++) {
			b = new BorderButton(new Point(1, j));
			b.addActionListener(borderAction);
			b.addMouseMotionListener(borderAction);
			b.addMouseListener(borderAction);
			supButtons.add(b);
		}

		//cases principales
		Cell c;
		for (int i = 0; i < sizeX; i++) {
			for (int j = 0; j < sizeY; j++) {
				c = new Cell(new Point(i, j), mouseAction, undoHandler, this);
				setCell(c, i, j);
				c.addKeyListener(keyAction);
			}
		}

		// initialisation de la s?lection
		tlSelectedCell = new Point(0, 0);
		brSelectedCell = new Point(0, 0);
		firstCell = getCell(0, 0);

		undoHandler.setActive(true);
		System.gc();
	}

	/*****************************************************************************
	 * Mise ? jour des comboBox de la ToolBar en fonction de firstCell.
	 *
	 * @pre - la toolbar doit exister
	 *		- firstCell doit exister
	 *		- undoHandler doit exister
	 * @post cbFont, cbSize, cbBackCol et cbTextCol sont mis ? jour en fonction
	 *			 de firstCell
	 */
	private void updateToolBar() {
		undoHandler.blockActiver(true);
		Font f = firstCell.getFont();
		cbFont.setSelectedItem(f.getName());
		cbSize.setSelectedItem(
			Integer.toString((int) Math.round(f.getSize() / 6)));
		cbTextCol.setSelectedItem(firstCell.getForeground());
		cbBackCol.setSelectedItem(firstCell.getBackground());
		undoHandler.blockActiver(false);
	}

	/*****************************************************************************
	 * Recr?? tout le panneau d'affichage cellPanel ? partir des ?l?ments 
	 * contenus dans la matrice cellGrid.
	 *
	 * @pre cellGrid doit ?tre cr??e et remplie d'objet BorderButton ou Cell
	 * @post cellPanel est recr?? et l'affichage est rafraichi
	 */
	private void redrawCellPanel() {
		if (cellPanel == null)
			cellPanel = new CellPanel();
		undoHandler.setActive(false);
		cellPanel.removeAll();
		System.gc();

		GridBagLayout gbl = new GridBagLayout();
		cellPanel.setLayout(gbl);
		cellPanel.revalidate();
		GridBagConstraints gbc = new GridBagConstraints();
		gbc.fill = GridBagConstraints.BOTH;
		gbc.weightx = 1.0;

		gbc.gridheight = 1;
		gbc.gridwidth = 1;
		// ATTENTION les indices du GBL sont d?cal?s de 1 par rapport au cellGrid

		// coin en haut ? gauche
		BorderButton b = (BorderButton) supButtons.get(0);
		gbc.gridy = 0;
		gbc.gridx = 0;
		gbl.setConstraints(b, gbc);
		cellPanel.add(b);

		//ligne de bordure gauche
		gbc.gridx = 0;
		for (int i = 0; i < getTabSize().height; i++) {
			b = (BorderButton) leftButtons.get(i + 1);
			gbc.gridy = i + 1;
			gbl.setConstraints(b, gbc);
			cellPanel.add(b);
		}
		//ligne de bordure sup?rieure
		gbc.gridy = 0;
		for (int j = 0; j < getTabSize().width; j++) {
			b = (BorderButton) supButtons.get(j + 1);
			gbc.gridx = j + 1;
			gbl.setConstraints(b, gbc);
			cellPanel.add(b);
		}

		//cases principales
		for (int i = 0; i < getTabSize().height; i++) {
			for (int j = 0; j < getTabSize().width; j++) {
				//ATTENTION, pour le gbl, y est la rang?e et x la colonne
				gbc.gridy = i + 1;
				gbc.gridx = j + 1;
				Cell c = getCell(i, j);
				if (c.getType() != CellInterf.TYPE_EMPTY) {
					gbc.gridheight = c.getCellSize().height;
					gbc.gridwidth = c.getCellSize().width;
					gbl.setConstraints(c, gbc);
					cellPanel.add(c);
				}
			}
		}
		cellPanel.revalidate();
		repaint();
		undoHandler.setActive(true);
	}

	/*****************************************************************************
	 * Convertion du sous-tableau en format HTML (sous forme de table dans 
	 * une cellule).
	 *
	 * @param CellArray ca : sous-tableau ? exporter
	 *		   int width, height : taille en pixel du cellPanel
	 * @return String: la repr?sentation HTML du sous-tableau ca
	 */
	private String exportSubTable(
		CellArray ca,
		int width,
		int height,
		int deepness) {
		deepness++;
		int h = ca.getBRCell().x - ca.getTLCell().x + 1;
		int w = ca.getBRCell().y - ca.getTLCell().y + 1;
		String s = "<td";
		if (h != 1)
			s += " rowspan=" + h;
		if (w != 1)
			s += " colspan=" + w;
		//s += " width=\"" + (int)Math.round((100.0 * taillecadre)/width ) + "%\"";
		//s += " height=\"" + (int)Math.round((100.0 * )/height) + "%\"";
		s += ">\n";
		s
			+= "<table border CELLPADDING=\"0\" CELLSPACING=\"0\" width=\"100%\" height=\"100%\" >\n";
		s += "<tr><td><table width=\"100%\" height=\"100%\">\n";

		for (int i = ca.getTLCell().x; i <= ca.getBRCell().x; i++) {
			s += "<tr valign=top>\n";
			for (int j = ca.getTLCell().y; j <= ca.getBRCell().y; j++) {
				Cell c = getCell(i, j);
				if (c.hasNextSubTable())
					s
						+= exportSubTable(
							c.nextSubTable(),
							width,
							height,
							deepness);
				else if (c.numSubTables() == deepness)
					s += c.toHTML(width, height);
			}
			s += "</tr>\n";
		}
		s += "</table></td></tr>\n</table>\n</td>\n";
		return s;
	}

	/*****************************************************************************
	 * Convertit un objet Color en code RGB ? utiliser en HTML
	 *
	 * @param Color col: la couleur ? convertir
	 * @return String: le code RGB
	 */
	private String getRGB(Color col) {
		String s;
		String c = Integer.toHexString(col.getRed());
		if (c.length() == 1)
			c = "0" + c;
		s = c;
		c = Integer.toHexString(col.getGreen());
		if (c.length() == 1)
			c = "0" + c;
		s += c;
		c = Integer.toHexString(col.getBlue());
		if (c.length() == 1)
			c = "0" + c;
		s += c;
		return s;
	}

	/*****************************************************************************
	 * Trouve la valeur de value dans le texte str, ce doit ?tre de la forme
	 * "value=qlqc". Renvoir le qlqc. Le qlqc se termine par un espace, par un tab,
	 * par un retour ? la ligne, par le symbole "plus petit" ("<") 
	 * ou "plus grand" (">"). Le qlqc ne peut donc pas contenir un de ces symboles.
	 *
	 * @param String value: le texte dont il faut trouver la valeur
	 * @param String str: texte dans lequel il faut chercher
	 * @return String: la valeur li?e ? value
	 */
	private String getValue(String value, String str) {
		int i = str.indexOf(value);
		int k = str.indexOf("=", i);
		String val = "";
		if (k != -1 && i != -1) {
			StringTokenizer st =
				new StringTokenizer(str.substring(k + 1), " \t\n><");
			val = st.nextToken();
		}
		return val;
	}

	/*****************************************************************************
	 * V?rifie si un objet o se trouve dans un tableau d'objet arr.
	 *
	 * @param Object[] arr: tableau dans lequel il faut chercher
	 * @param Object o: objet ? chercher dans arr
	 * @return boolean: true si o se trouve dans arr
	 */
	private boolean arrayContains(Object[] arr, Object o) {
		for (int i = 0; i < arr.length; i++) {
			if (arr[i].equals(o))
				return true;
		}
		return false;
	}

	/*****************************************************************************
	 * Mofifie le style d'une case en fonction de la s?lection du style de
	 * la Toolbar (cbStyle).
	 *
	 * @param Cell cell: cellules ? modifier
	 * @post cell a le m?me style que cbStyle
	 */
	private void setCellStyle(Cell cell) {
		undoHandler.setActive(false);

		// Style
		int i = 0;
		while (i < styles.length) {
			if (cbStyle.getSelectedItem().equals(styles[i][0]))
				break;
			i++;
		}

		// Police + taille + gras + italique
		if (styles[i][5].equals("O"))
			cell.setBold(true);
		else
			cell.setBold(false);
		if (styles[i][6].equals("O"))
			cell.setItalic(true);
		else
			cell.setItalic(false);
		if (styles[i][7].equals("O"))
			cell.setUnderline(true);
		else
			cell.setUnderline(false);
		// 6 : m?me valeur que dans l'?diteur de texte. Mais PQ cette valeur ??
		Font f =
			new Font(
				styles[i][1],
				Font.PLAIN,
				(int) Math.round(6 * Integer.parseInt(styles[i][2])));

		// Couleur du texte
		Color c = Color.black;
		if (styles[i][3].equals("Vert")) {
			c = Color.green;
		} else if (styles[i][3].equals("Bleu")) {
			c = Color.blue;
		} else if (styles[i][3].equals("Jaune")) {
			c = Color.yellow;
		} else if (styles[i][3].equals("Gris")) {
			c = Color.gray;
		} else if (styles[i][3].equals("Rouge")) {
			c = Color.red;
		} else if (styles[i][3].equals("Blanc")) {
			c = Color.white;
		}

		// Alignement
		int align = StyleConstants.ALIGN_LEFT;
		if (styles[i][4].equals("C")) {
			align = StyleConstants.ALIGN_CENTER;
		} else if (styles[i][4].equals("R")) {
			align = StyleConstants.ALIGN_RIGHT;
		}

		cell.setFont(f);
		cell.setHorizontalAlignment(align);
		cell.setForeground(c);
		undoHandler.setActive(true);

	}

	/*****************************************************************************
	 * Fusion de toutes les cellules comprises entre tlC et brC, elles
	 * sont toutes (tlC comprise) remplac?es par une "EmptyCell" pointant
	 * vers la position tlC.
	 *
	 * @param Point tlC, brC: tlC > (0,0) et brC > (0,0)	
	 * @post toutes les cellules comprises entre tlC et brC sont remplac?es
	 *			 par des cellules de type TYPE_EMPTY pointant vers tlC
	 */
	private void fusionCells(Point tlC, Point brC) {
		for (int i = tlC.x; i <= brC.x; i++) {
			for (int j = tlC.y; j <= brC.y; j++) {
				setCell(new Cell(new Point(tlC)), i, j);
			}
		}
	}

	/*****************************************************************************
	 * Cr?ation d'un sous-tableau allant de la position tl ? la position br.
	 * Si il y a incompatibilit? avec un autre sous-tableau (intersection
	 * des bords ou s'ils sont exactement les m?me), alors il n'y a pas de 
	 * modifications effectu?es.
	 *
	 * @param Point tl, br: tl > (0,0) et br > (0,0)
	 * @return boolean: true si modification effectu?e, false si incompatibilit?
	 * @post le tableau contient un sous-tableau de tl ? tc si compatibilit?
	 */
	private boolean createSubTable(Point tl, Point br) {
		Iterator iter = subTableList.iterator();
		while (iter.hasNext()) {
			CellArray ca = (CellArray) iter.next();
			if (!ca.isCompatibleWith(tl, br, false)) {
				System.out.println(
					"You can't put a border there, "
						+ "you can't cross borders or fusionned cells.");
				return false;
			}
		}

		// permet de g?rer toute la modif en un seul undo
		CompoundEdit ce = new CompoundEdit();
		StateEdit edit = new StateEdit(new SubTableStateEdit());
		CellArray ca = new CellArray(tl, br);
		subTableList.add(ca);
		edit.end();
		ce.addEdit(edit);

		for (int i = tl.x; i <= br.x; i++) {
			for (int j = tl.y; j <= br.y; j++) {
				edit = new StateEdit(getCell(i, j));
				getCell(i, j).putInSubTable(ca);
				edit.end();
				ce.addEdit(edit);
			}
		}
		ce.end();
		undoHandler.addEdit(ce);
		return true;
	}

	/*****************************************************************************
	 * Suppresion du sous-tableau ca.
	 *
	 * @param CellArray ca: le sous-tableau ? supprimer
	 * @post le tableau ca a ?t? supprim? du tableur
	 */
	private void removeSubTable(CellArray ca) {
		CompoundEdit ce = new CompoundEdit();
		StateEdit edit = new StateEdit(new SubTableStateEdit());
		subTableList.remove(ca);
		edit.end();
		ce.addEdit(edit);

		for (int i = ca.getTLCell().x; i <= ca.getBRCell().x; i++) {
			for (int j = ca.getTLCell().y; j <= ca.getBRCell().y; j++) {
				edit = new StateEdit(getCell(i, j));
				getCell(i, j).removeFromSubTable(ca);
				edit.end();
				ce.addEdit(edit);
			}
		}
		ce.end();
		undoHandler.addEdit(ce);
	}

	/*****************************************************************************
	 * Modification de la largeur d'une colonne enti?re du tableau.
	 *
	 * @param int col: col > 0 : colonne ? modifier
	 * @param int size: size > 0 : taille ? rajouter ? la colonne 
	 * @post la colonne col a une largeur size
	 */
	private void setColSize(int col, int size) {
		for (int i = 0; i < getTabSize().height; i++) {
			//permet de r?cup?rer la bonne case, si la case en p est TYPE_EMPTY
			Point p = getCell(i, col).getLocation();
			Cell c = getCell(p);
			int width = c.getSize().width + size;
			c.setPreferredSize(new Dimension(width, c.getSize().height));
		}
	}

	/*****************************************************************************
	 * Modification de la hauteur d'une ligne enti?re du tableau.
	 *
	 * @param int row: row > 0 : ligne ? modifier
	 * @param int size: size > 0 : taille ? rajouter ? la ligne 
	 * @post la ligne row a une hauteur size
	 */
	void setRowSize(int row, int size) {
		// il faut d'abords v?rifier qu'on peut diminuer la taille des cases
		if (size < 0) {
			for (int j = 0; j < getTabSize().width; j++) {
				Point p = getCell(row, j).getLocation();
				Cell c = getCell(p);
				if (c.getSize().height + size < c.getMinHeight())
					return;
			}
		}
		for (int j = 0; j < getTabSize().width; j++) {
			Point p = getCell(row, j).getLocation();
			Cell c = getCell(p);
			int height = c.getSize().height + size;
			c.setPreferredSize(new Dimension(c.getSize().width, height));
		}
	}

	/*****************************************************************************
	 * Met ? jour la taille des cases pour qu'on puisse voir tout le contenu 
	 * de chaque case.
	 *
	 * @post la hauteur des cases est modifi?e si n?cessaire
	 */
	private void updateCellHeight() {
		for (int i = 0; i < getTabSize().height; i++) {
			for (int j = 0; j < getTabSize().width; j++) {
				getCell(i, j).updateHeight();
			}
		}
	}

	/*****************************************************************************
	 * Rajoute une ligne au tableau, ? la position x.
	 * 
	 * @param int x, y : int x <= getTabSize().height, position d'insertion
	 * 		   boolean insert: true si on ins?re une ligne
	 * @post une ligne remplie de cases vides est rajout?e ? la position i 
	 */
	private void addRow(int x, boolean insert) {
		// rajout du bouton
		if (leftButtons.size() - 1 < getTabSize().height + 1) {
			BorderButton b =
				new BorderButton(new Point(0, getTabSize().height));
			b.addActionListener(borderAction);
			b.addMouseMotionListener(borderAction);
			b.addMouseListener(borderAction);
			b.addKeyListener(keyAction);
			leftButtons.add(b);
		}
		CompoundEdit edit = new RedrawingCompoundEdit();
		StateEdit se = new StateEdit(cellGrid);
		// rajout de la ligne dans le tableau
		cellGrid.insertRow(x);

		for (int j = 0; j < getTabSize().width; j++) {
			Cell c = new Cell(new Point(x, j), mouseAction, undoHandler, this);
			c.setPreferredSize(new Dimension(((BorderButton)supButtons.get(j+1)).getSize().width, 
												 ((BorderButton)leftButtons.get(x+1)).getSize().height));
			setCell(c, x, j);
			c.addKeyListener(keyAction);
			setCellStyle(c);
		}

		// mise ? jour des positions
		for (int i = x + 1; i < getTabSize().height; i++) {
			for (int j = 0; j < getTabSize().width; j++) {
				Cell c = getCell(i, j);
				/* si c'est une case vide => position=position de la case principale
				 * => la redescendre que si celle-ci est plus bas ou en x
				 */
				if (c.getType() != CellInterf.TYPE_EMPTY
					|| (c.getType() == CellInterf.TYPE_EMPTY
						&& c.getLocation().x >= x)) {
					Point p = c.getLocation();
					StateEdit se2 = new StateEdit(c);
					p.translate(1, 0);
					c.setLocation(p);
					se2.end();
					edit.addEdit(se2);
				}
			}
		}
		//gestion des cases fusionn?es:
		if (x + 1 < getTabSize().height)
			for (int j = 0; j < getTabSize().width; j++) {
				Cell c = getCell(x + 1, j);
				// 2e condition pour ne traiter qu'une seule fois une case
				// plus large qu'une seule colonne
				if (c.getType() == CellInterf.TYPE_EMPTY) {
					if (c.getLocation().y == j) {
						Cell mast = getCell(c.getLocation());
						Dimension dim = mast.getCellSize();
						StateEdit se2 = new StateEdit(mast);
						dim.setSize(dim.width, dim.height + 1);
						mast.setCellSize(dim);
						se2.end();
						edit.addEdit(se2);
					}
					// si x = x+1, on a ins?r? une case au dessus du tableau => 
					// il ne faut pas la transformer en case vide
					if (c.getLocation().x != x + 1)
						setCell(new Cell(c.getLocation()), x, j);
				}
			}

		se.end();
		edit.addEdit(se);

		//gestion des encadrements:
		Iterator iter = subTableList.iterator();
		while (iter.hasNext()) {
			CellArray ca = (CellArray) iter.next();
			Point tl = ca.getTLCell();
			Point br = ca.getBRCell();
			se = new StateEdit(ca);

			if (tl.x >= x) {
				tl.translate(1, 0);
				ca.setTLCell(tl);
			}

			if (br.x >= x && insert) {
				br.translate(1, 0);
				ca.setBRCell(br);
				//si les nouvelles cases sont dans un tableau, il faut les mettre dedant
				if (tl.x <= x)
					for (int j = tl.y; j <= br.y; j++) {
						Cell c = getCell(x, j);
						StateEdit se2 = new StateEdit(c);
						c.putInSubTable(ca);
						se2.end();
						edit.addEdit(se2);
					}
			}
			if (x == tl.x) {
				// il faut que l'ancienne premi?re case ne soit plus la "master"
				// au cas o? on change la premi?re ligne
				tl.translate(1, 0); // l'ancienne master a descendu de 1 case.
				Cell c = getCell(tl);
				StateEdit se2 = new StateEdit(c);
				c.removeFromSubTable(ca);
				// il faut que l'ancienne premi?re case soit remise dans le tableau
				c.putInSubTable(ca);
				se2.end();
				edit.addEdit(se2);
			}
			se.end();
			edit.addEdit(se);
		}

		edit.end();
		undoHandler.addEdit(edit);

		firstCell = getCell(tlSelectedCell);
		redrawCellPanel();
		firstCell.grabFocus();
	}
	/*****************************************************************************
	 * Efface la ligne x du tableau.
	 * 
	 * @param int x : int x < getTabSize().height, position de suppression
	 * @post la ligne x est suppprim?e.
	 */
	private void delRow(int x) {
		// il ne faut pas effacer le bouton, sinon le undo va foirer!
		// une autre solution est de tenir compte des marges pour le undo-redo...

		CompoundEdit edit = new RedrawingCompoundEdit();

		// gestion de la fusion	: n?cessaire sinon on perd de l'info!
		// uniquement si on efface une case maitresse (Top-Left)
		StateEdit se = new StateEdit(cellGrid);

		for (int j = 0; j < getTabSize().width; j++) {
			Cell c = getCell(x, j);
			// on efface la master d'une cellule fusionn?e => la suivante doit 
			// devenir master
			if (c.getCellSize().height > 1) {
				Dimension dim = c.getCellSize();
				dim.setSize(dim.width, dim.height - 1);
				setCell(new Cell(new Point(x + 1, j),
				/* +1 car c'est la position actuelle de cette cellule*/
				dim, mouseAction, undoHandler, this), x + 1, j);
				setCellStyle(getCell(x + 1, j));
			} else if (
				// sinon, on diminue la hauteur de la master de 1
			c.getType()
				== CellInterf.TYPE_EMPTY
					&& j == c.getLocation().y) {
				Cell mast = getCell(c.getLocation());
				Dimension dim = mast.getCellSize();
				StateEdit se2 = new StateEdit(mast);
				dim.setSize(dim.width, dim.height - 1);
				mast.setCellSize(dim);
				se2.end();
				edit.addEdit(se2);
			}
		}

		// effacement de la ligne dans le tableau
		cellGrid.delRow(x);

		se.end();
		edit.addEdit(se);

		// mise ? jour des positions
		for (int i = x; i < getTabSize().height; i++) {
			for (int j = 0; j < getTabSize().width; j++) {
				Cell c = getCell(i, j);
				/* si c'est une case vide => position=position de la case principale
				 * => la remonter que si celle-ci est plus bas ou en x
				 */
				if (c.getType() != CellInterf.TYPE_EMPTY
					|| (c.getType() == CellInterf.TYPE_EMPTY
						&& c.getLocation().x > x)) {
					Point p = c.getLocation();
					StateEdit se2 = new StateEdit(c);
					p.translate(-1, 0);
					c.setLocation(p);
					se2.end();
					edit.addEdit(se2);
				}
			}
		}

		se = new StateEdit(new SubTableStateEdit());
		//gestion des encadrements:
		Iterator iter = subTableList.iterator();
		while (iter.hasNext()) {
			CellArray ca = (CellArray) iter.next();
			Point tl = ca.getTLCell();
			Point br = ca.getBRCell();
			StateEdit se2 = new StateEdit(ca);

			if (tl.x == x && br.x == x) {
				iter.remove();
			} else {
				if (tl.x > x && tl.x != 0) {
					tl.translate(-1, 0);
					ca.setTLCell(tl);
				}
				if (br.x >= x) {
					br.translate(-1, 0);
					ca.setBRCell(br);
				}
			}
			se2.end();
			edit.addEdit(se2);

			if (x == tl.x) {
				// il faut que la nouvelle premi?re case devienne la "master"
				// au cas o? on change la premi?re ligne
				Cell c = getCell(tl);
				StateEdit se3 = new StateEdit(c);
				c.removeFromSubTable(ca);
				// il faut que la nouvelle premi?re case soit remise dans le tableau
				c.putInSubTable(ca);
				se3.end();
				edit.addEdit(se3);
			}
		}
		se.end();
		edit.addEdit(se);

		edit.end();
		undoHandler.addEdit(edit);

		if (brSelectedCell.x >= getTabSize().height)
			brSelectedCell.translate(-1, 0);

		if (tlSelectedCell.x >= getTabSize().height) {
			tlSelectedCell.translate(-1, 0);
			brSelectedCell = new Point(tlSelectedCell);
		}
		firstCell = getCell(tlSelectedCell);
		redrawCellPanel();
		firstCell.grabFocus();
	}

	/*****************************************************************************
	 * Rajoute une colonne au tableau, ? la position y.
	 * 
	 * @param int x, y : int y <= getTabSize().width, position d'insertion
	 * 		   boolean insert: true si on ins?re une colonne
	 * @post une colonne remplie de cases vides est rajout?e ? la position i 
	 */
	private void addCol(int y, boolean insert) {
		// rajout du bouton
		if (supButtons.size() - 1 < getTabSize().width + 1) {
			BorderButton b = new BorderButton(new Point(1, getTabSize().width));
			b.addActionListener(borderAction);
			b.addMouseMotionListener(borderAction);
			b.addMouseListener(borderAction);
			b.addKeyListener(keyAction);
			supButtons.add(b);
		}
		CompoundEdit edit = new RedrawingCompoundEdit();
		StateEdit se = new StateEdit(cellGrid);
		// rajout de la ligne dans le tableau
		cellGrid.insertCol(y);

		for (int i = 0; i < getTabSize().height; i++) {
			Cell c = new Cell(new Point(i, y), mouseAction, undoHandler, this);
			setCell(c, i, y);
			c.addKeyListener(keyAction);
			setCellStyle(c);
		}

		// mise ? jour des positions
		for (int i = 0; i < getTabSize().height; i++) {
			for (int j = y + 1; j < getTabSize().width; j++) {
				Cell c = getCell(i, j);
				/* si c'est une case vide => position=position de la case principale
				 * => la redressendre que si celle-ci est plus bas ou en x
				 */
				if (c.getType() != CellInterf.TYPE_EMPTY
					|| (c.getType() == CellInterf.TYPE_EMPTY
						&& c.getLocation().y >= y)) {
					Point p = c.getLocation();
					StateEdit se2 = new StateEdit(c);
					p.translate(0, 1);
					c.setLocation(p);
					se2.end();
					edit.addEdit(se2);
				}
			}
		}
		//gestion des cases fusionn?es:
		if (y + 1 < getTabSize().width)
			for (int i = 0; i < getTabSize().height; i++) {
				Cell c = getCell(i, y + 1);
				// 2e condition pour ne traiter qu'une seule fois une case
				// plus large qu'une seule colonne
				if (c.getType() == CellInterf.TYPE_EMPTY) {
					if (c.getLocation().x == i) {
						Cell mast = getCell(c.getLocation());
						Dimension dim = mast.getCellSize();
						StateEdit se2 = new StateEdit(mast);
						dim.setSize(dim.width + 1, dim.height);
						mast.setCellSize(dim);
						se2.end();
						edit.addEdit(se2);
					}
					// si y = y+1, on a ins?r? une case ? gauche du tableau => 
					// il ne faut pas la transformer en case vide
					if (c.getLocation().y != y + 1)
						setCell(new Cell(c.getLocation()), i, y);
				}
			}
		se.end();
		edit.addEdit(se);

		//gestion des encadrements:
		Iterator iter = subTableList.iterator();
		while (iter.hasNext()) {
			CellArray ca = (CellArray) iter.next();
			Point tl = ca.getTLCell();
			Point br = ca.getBRCell();
			se = new StateEdit(ca);

			if (tl.y >= y) {
				tl.translate(0, 1);
				ca.setTLCell(tl);
			}

			if (br.y >= y && insert) {
				br.translate(0, 1);
				ca.setBRCell(br);
				//si les nouvelles cases sont dans un tableau, il faut les mettre dedant
				if (tl.y <= y)
					for (int i = tl.x; i <= br.x; i++) {
						Cell c = getCell(i, y);
						StateEdit se2 = new StateEdit(c);
						c.putInSubTable(ca);
						se2.end();
						edit.addEdit(se2);
					}
			}
			if (y == tl.y) {
				// il faut que l'ancienne premi?re case ne soit plus la "master"
				// au cas o? on change la premi?re ligne
				tl.translate(0, 1); // l'ancienne master a descendu de 1 case.
				Cell c = getCell(tl);
				StateEdit se2 = new StateEdit(c);
				c.removeFromSubTable(ca);
				// il faut que l'ancienne premi?re case soit remise dans le tableau
				c.putInSubTable(ca);
				se2.end();
				edit.addEdit(se2);
			}
			se.end();
			edit.addEdit(se);
		}

		edit.end();
		undoHandler.addEdit(edit);

		firstCell = getCell(tlSelectedCell);
		redrawCellPanel();
		firstCell.grabFocus();
	}
	/*****************************************************************************
	 * Efface la colonne y du tableau.
	 * 
	 * @param int y : int y < getTabSize().width, position de suppression
	 * @post la ligne y est suppprim?e.
	 */
	private void delCol(int y) {
		// il ne faut pas effacer le bouton, sinon le undo va foirer!
		// une autre solution est de tenir compte des marges pour le undo-redo...

		CompoundEdit edit = new RedrawingCompoundEdit();

		// gestion de la fusion	: n?cessaire sinon on perd de l'info!
		// uniquement si on efface une case maitresse (Top-Left)
		StateEdit se = new StateEdit(cellGrid);

		for (int i = 0; i < getTabSize().height; i++) {
			Cell c = getCell(i, y);
			if (c.getCellSize().width > 1) {
				Dimension dim = c.getCellSize();
				dim.setSize(dim.width - 1, dim.height);
				setCell(new Cell(new Point(i, y + 1),
				/* +1 car c'est la position actuelle de cette cellule*/
				dim, mouseAction, undoHandler, this), i, y + 1);
				setCellStyle(getCell(i, y + 1));
			} else if (
				c.getType() == CellInterf.TYPE_EMPTY
					&& i == c.getLocation().x) {
				Cell mast = getCell(c.getLocation());
				Dimension dim = mast.getCellSize();
				StateEdit se2 = new StateEdit(mast);
				dim.setSize(dim.width - 1, dim.height);
				mast.setCellSize(dim);
				se2.end();
				edit.addEdit(se2);
			}
		}

		// effacement de la colonne dans le tableau
		cellGrid.delCol(y);

		se.end();
		edit.addEdit(se);

		// mise ? jour des positions
		for (int i = 0; i < getTabSize().height; i++) {
			for (int j = y; j < getTabSize().width; j++) {
				Cell c = getCell(i, j);
				/* si c'est une case vide => position=position de la case principale
				 * => la remonter que si celle-ci est plus bas ou en x
				 */
				if (c.getType() != CellInterf.TYPE_EMPTY
					|| (c.getType() == CellInterf.TYPE_EMPTY
						&& c.getLocation().y > y)) {
					Point p = c.getLocation();
					StateEdit se2 = new StateEdit(c);
					p.translate(0, -1);
					c.setLocation(p);
					se2.end();
					edit.addEdit(se2);
				}
			}
		}

		se = new StateEdit(new SubTableStateEdit());
		//gestion des encadrements:
		Iterator iter = subTableList.iterator();
		while (iter.hasNext()) {
			CellArray ca = (CellArray) iter.next();
			Point tl = ca.getTLCell();
			Point br = ca.getBRCell();
			StateEdit se2 = new StateEdit(ca);

			if (tl.y == y && br.y == y) {
				iter.remove();
			} else {
				if (tl.y > y && tl.y != 0) {
					tl.translate(0, -1);
					ca.setTLCell(tl);
				}
				if (br.y >= y) {
					br.translate(0, -1);
					ca.setBRCell(br);
				}
			}
			se2.end();
			edit.addEdit(se2);

			if (y == tl.y) {
				// il faut que la nouvelle premi?re case devienne la "master"
				// au cas o? on change la premi?re ligne
				Cell c = getCell(tl);
				StateEdit se3 = new StateEdit(c);
				c.removeFromSubTable(ca);
				// il faut que la nouvelle premi?re case soit remise dans le tableau
				c.putInSubTable(ca);
				se3.end();
				edit.addEdit(se3);
			}
		}
		se.end();
		edit.addEdit(se);

		edit.end();
		undoHandler.addEdit(edit);

		if (brSelectedCell.y >= getTabSize().width)
			brSelectedCell.translate(0, -1);

		if (tlSelectedCell.y >= getTabSize().width) {
			tlSelectedCell.translate(0, -1);
			brSelectedCell = new Point(tlSelectedCell);
		}
		firstCell = getCell(tlSelectedCell);
		redrawCellPanel();
		firstCell.grabFocus();
	}

	//******************* Classes internes priv?es *******************************

	/*****************************************************************************
	 * Classe interne g?rant les actions de la souris sur les cellules du tableur.
	 * Elle g?re la s?lection de une ou plusieurs cases.
	 */
	private class MouseAction extends MouseAdapter {
		//true si l'utilisateur est en train de s?lectionner (bouton enfonc?) 
		private boolean selectionning = false;

		public void mousePressed(MouseEvent e) {
			/* 
			 * S?lection de la case o? on a appui? sur la souris et passage en mode 
			 * s?lectionning
			 */
			Cell c = (Cell) e.getSource();

			int i = c.getLocation().x;
			int j = c.getLocation().y;

			if (e.getButton() == MouseEvent.BUTTON2)
				/*externalListener.println(
					"("
						+ i
						+ ","
						+ j
						+ "): "
						+ getCell(i, j).getLocation()
						+ "  "
						+ getCell(i, j).getSize()
						+ " "
						+ getCell(i, j).getType())*/;

			firstCell = c;
			tlSelectedCell.setLocation(c.getLocation());
			brSelectedCell.setLocation(c.getLocation());
			repaint();
			c.grabFocus();
			updateToolBar();

			if (e.getButton() == MouseEvent.BUTTON1)
				selectionning = true;
		}

		public void mouseEntered(MouseEvent e) {
			updateCellHeight();

			/*
			 * Si on est en mode s?lectionning, il faut trouver les deux valeurs 
			 * extr?mes s?lectionn?es Top-Left et Bottom-Right), tout en gardant comme
			 * firstCell celle sur laquelle on a cliqu? au d?part.
			 */
			if (!selectionning)
				return;
			Cell cell = (Cell) e.getSource();

			Point xy1 = firstCell.getLocation();
			Dimension hw1 = firstCell.getCellSize();
			Point xy2 = cell.getLocation();
			Dimension hw2 = cell.getCellSize();
			// en x :
			int top;
			int bottom;
			top = (xy1.x <= xy2.x ? xy1.x : xy2.x);
			bottom =
				(xy1.x + hw1.height >= xy2.x + hw2.height
					? xy1.x + hw1.height - 1
					: xy2.x + hw2.height - 1);
			// en y :
			int left;
			int right;
			left = (xy1.y <= xy2.y ? xy1.y : xy2.y);
			right =
				(xy1.y + hw1.width >= xy2.y + hw2.width
					? xy1.y + hw1.width - 1
					: xy2.y + hw2.width - 1);

			/* algorithme it?ratif permettant de ne pas couper des cellules fusionn?es
			 * lors de la s?lection et donc d'agrandir malgr? l'utilisateur la 
			 * zone s?lectionn?e pour inclure l'enti?ret? des cellules.
			 */
			boolean isModified = true;
			Cell c;
			while (isModified) {
				isModified = false;
				for (int i = top; i <= bottom; i++) {
					for (int j = left; j <= right; j++) {
						c = getCell(i, j);
						// en x
						if (c.getLocation().x < top) {
							top = c.getLocation().x;
							isModified = true;
						}
						if (c.getLocation().x + c.getCellSize().height - 1
							> bottom) {
							bottom =
								c.getLocation().x + c.getCellSize().height - 1;
							isModified = true;
						}
						// en y
						if (c.getLocation().y < left) {
							left = c.getLocation().y;
							isModified = true;
						}
						if (c.getLocation().y + c.getCellSize().width - 1
							> right) {
							right =
								c.getLocation().y + c.getCellSize().width - 1;
							isModified = true;
						}
					}
				}
			}
			tlSelectedCell.setLocation(top, left);
			brSelectedCell.setLocation(bottom, right);
			repaint();
		}
		/*
		 * Quand on lache la souris, on sort du mode selectionning.
		 */
		public void mouseReleased(MouseEvent e) {
			selectionning = false;
		}
	}

	/*****************************************************************************
	 * Classe interne g?rant les actions sur les boutons de la ToolBar. A chaque
	 * bouton correspond une m?thode ex?cut?e. Le itemEvnet est n?cessaire car
	 * ca permet aux JComboBox de ne pas renvoyer un Event lorsque l'item 
	 * s?lectionn? n'est pas chang? (lors d'un updateToolBar).
	 */
	private class ToolBarAction implements ActionListener, ItemListener {
		public void actionPerformed(ActionEvent e) {
			Object o = e.getSource();
			if (o == bBold)
				setBold();
			else if (o == bItalic)
				setItalic();
			else if (o == bUnderline)
				setUnderline();
			else if (o == bLeft)
				setLeft();
			else if (o == bCenter)
				setCenter();
			else if (o == bRight)
				setRight();
			else if (o == bFusion)
				fusion();
			else if (o == bUndo)
				undo();
			else if (o == bRedo)
				redo();
			else if (o == bBorder)
				border();
			else if (o == bDebug)
				debug();
		}

		public void itemStateChanged(ItemEvent e) {
			if (e.getStateChange() != ItemEvent.SELECTED)
				return;

			Object o = e.getSource();
			if (o == cbStyle)
				setStyle();
			else if (o == cbFont)
				setFont();
			else if (o == cbSize)
				setSize();
			else if (o == cbTextCol)
				setTextColor();
			else if (o == cbBackCol)
				setBackColor();
		}

		private void debug() {
			/*Iterator iter = subTableList.iterator();
			for (int i = 0; i < getTabSize().height; i++) {
				for (int j = 0; j < getTabSize().width; j++) {
					getCell(i, j).initExport();
				}
			}
			
			while (iter.hasNext()) {
				CellArray ca = (CellArray) iter.next();
				System.out.println(
					"ca: TL: " + ca.getTLCell() + " BR: " + ca.getBRCell());
			}
			for (int i = 0; i < getTabSize().height; i++) {
				for (int j = 0; j < getTabSize().width; j++) {
					Cell c = getCell(i, j);
					System.out.println(
						"cell("
							+ i
							+ ","
							+ j
							+ "): num:"
							+ c.numSubTables()
							+ c.hasNextSubTable());
				}
			}
			System.out.println("");
			*/

			/*externalListener.println("");*/

			/*for (int i = 0; i < getTabSize().height; i++) {
				for (int j = 0; j < getTabSize().width; j++) {
					System.out.println(
						"("
							+ i
							+ ","
							+ j
							+ "): "
							+ getCell(i, j).getLocation()
							+ "  "
							+ getCell(i, j).getCellSize()
							+ " "
							+ getCell(i, j).getType());
				}
			}

			System.out.println("");*/
		}

		private void setStyle() {
			CompoundEdit ce = new CompoundEdit();
			for (int i = tlSelectedCell.x; i <= brSelectedCell.x; i++) {
				for (int j = tlSelectedCell.y; j <= brSelectedCell.y; j++) {
					Cell cell = getCell(i, j);
					StateEdit edit = new StateEdit(cell);
					setCellStyle(cell);
					edit.end();
					ce.addEdit(edit);
					cell.updateHeight();
				}
			}
			ce.end();
			undoHandler.addEdit(ce);
			updateToolBar();
		}

		private void setFont() {
			undoHandler.setActive(false);
			CompoundEdit ce = new CompoundEdit();
			for (int i = tlSelectedCell.x; i <= brSelectedCell.x; i++) {
				for (int j = tlSelectedCell.y; j <= brSelectedCell.y; j++) {
					Cell cell = getCell(i, j);
					Font f = cell.getFont();
					StateEdit edit = new StateEdit(cell);
					cell.setFont(
						new Font(
							(String) cbFont.getSelectedItem(),
							f.getStyle(),
							f.getSize()));
					edit.end();
					ce.addEdit(edit);
					cell.updateHeight();
				}
			}
			ce.end();
			undoHandler.setActive(true);
			undoHandler.addEdit(ce);
		}

		private void setSize() {
			undoHandler.setActive(false);
			CompoundEdit ce = new CompoundEdit();
			for (int i = tlSelectedCell.x; i <= brSelectedCell.x; i++) {
				for (int j = tlSelectedCell.y; j <= brSelectedCell.y; j++) {
					Cell cell = getCell(i, j);
					Font f = cell.getFont();
					StateEdit edit = new StateEdit(cell);
					cell.setFont(
						f.deriveFont(
							(float) 6
								* Integer.parseInt(
									(String) cbSize.getSelectedItem())));
					cell.updateHeight();
					edit.end();
					ce.addEdit(edit);
				}
			}
			ce.end();
			UndoHandler erf = undoHandler;
			undoHandler.setActive(true);
			undoHandler.addEdit(ce);
			firstCell.grabFocus();
		}

		private void setBold() {
			undoHandler.setActive(false);
			CompoundEdit ce = new CompoundEdit();
			for (int i = tlSelectedCell.x; i <= brSelectedCell.x; i++) {
				for (int j = tlSelectedCell.y; j <= brSelectedCell.y; j++) {
					Cell cell = getCell(i, j);
					StateEdit edit = new StateEdit(cell);
					cell.setBold();
					edit.end();
					ce.addEdit(edit);
					cell.updateHeight();
				}
			}
			ce.end();
			undoHandler.setActive(true);
			undoHandler.addEdit(ce);
		}

		private void setItalic() {
			undoHandler.setActive(false);
			CompoundEdit ce = new CompoundEdit();
			for (int i = tlSelectedCell.x; i <= brSelectedCell.x; i++) {
				for (int j = tlSelectedCell.y; j <= brSelectedCell.y; j++) {
					Cell cell = getCell(i, j);
					StateEdit edit = new StateEdit(cell);
					cell.setItalic();
					edit.end();
					ce.addEdit(edit);
					cell.updateHeight();
				}
			}
			ce.end();
			undoHandler.setActive(true);
			undoHandler.addEdit(ce);
		}

		private void setUnderline() {
			undoHandler.setActive(false);
			CompoundEdit ce = new CompoundEdit();
			for (int i = tlSelectedCell.x; i <= brSelectedCell.x; i++) {
				for (int j = tlSelectedCell.y; j <= brSelectedCell.y; j++) {
					Cell cell = getCell(i, j);
					StateEdit edit = new StateEdit(cell);
					cell.setUnderline();
					edit.end();
					ce.addEdit(edit);
					cell.updateHeight();
				}
			}
			ce.end();
			undoHandler.setActive(true);
			undoHandler.addEdit(ce);
		}

		private void setTextColor() {
			undoHandler.setActive(false);
			CompoundEdit ce = new CompoundEdit();
			for (int i = tlSelectedCell.x; i <= brSelectedCell.x; i++) {
				for (int j = tlSelectedCell.y; j <= brSelectedCell.y; j++) {
					Cell cell = getCell(i, j);
					StateEdit edit = new StateEdit(cell);
					cell.setForeground((Color) cbTextCol.getSelectedItem());
					edit.end();
					ce.addEdit(edit);
				}
			}
			ce.end();
			undoHandler.setActive(true);
			undoHandler.addEdit(ce);
		}

		private void setBackColor() {
			undoHandler.setActive(false);
			CompoundEdit ce = new CompoundEdit();
			for (int i = tlSelectedCell.x; i <= brSelectedCell.x; i++) {
				for (int j = tlSelectedCell.y; j <= brSelectedCell.y; j++) {
					Cell cell = getCell(i, j);
					StateEdit edit = new StateEdit(cell);
					cell.setBackground((Color) cbBackCol.getSelectedItem());
					edit.end();
					ce.addEdit(edit);
				}
			}
			ce.end();
			undoHandler.setActive(true);
			undoHandler.addEdit(ce);
		}

		private void setLeft() {
			undoHandler.setActive(false);
			CompoundEdit ce = new CompoundEdit();
			for (int i = tlSelectedCell.x; i <= brSelectedCell.x; i++) {
				for (int j = tlSelectedCell.y; j <= brSelectedCell.y; j++) {
					Cell cell = getCell(i, j);
					cell.setUnderline(false);
					StateEdit edit = new StateEdit(cell);
					cell.setHorizontalAlignment(StyleConstants.ALIGN_LEFT);
					edit.end();
					ce.addEdit(edit);
				}
			}
			ce.end();
			undoHandler.setActive(true);
			undoHandler.addEdit(ce);
		}

		private void setCenter() {
			undoHandler.setActive(false);
			CompoundEdit ce = new CompoundEdit();
			for (int i = tlSelectedCell.x; i <= brSelectedCell.x; i++) {
				for (int j = tlSelectedCell.y; j <= brSelectedCell.y; j++) {
					Cell cell = getCell(i, j);
					StateEdit edit = new StateEdit(cell);
					cell.setHorizontalAlignment(StyleConstants.ALIGN_CENTER);
					edit.end();
					ce.addEdit(edit);
				}
			}
			ce.end();
			undoHandler.setActive(true);
			undoHandler.addEdit(ce);
		}

		private void setRight() {
			undoHandler.setActive(false);
			CompoundEdit ce = new CompoundEdit();
			for (int i = tlSelectedCell.x; i <= brSelectedCell.x; i++) {
				for (int j = tlSelectedCell.y; j <= brSelectedCell.y; j++) {
					Cell cell = getCell(i, j);
					StateEdit edit = new StateEdit(cell);
					cell.setHorizontalAlignment(StyleConstants.ALIGN_RIGHT);
					edit.end();
					ce.addEdit(edit);
				}
			}
			ce.end();
			undoHandler.setActive(true);
			undoHandler.addEdit(ce);
		}

		private void fusion() {
			if (tlSelectedCell.equals(brSelectedCell))
				return;
			Iterator iter = subTableList.iterator();
			/* il faut v?rifier qu'on ne veut pas fusionner des cellules qui ne sont
			 * pas dans les m?me sous-tableau ni fusionner un groupe de cellules dont 
			 * certaines sont encadr?es.
			 */
			while (iter.hasNext()) {
				CellArray ca = (CellArray) iter.next();
				if (!ca
					.isCompatibleWith(tlSelectedCell, brSelectedCell, true)) {
					System.out.println(
						"You can't fusion these cells, there are not in the same subtables.");
					return;
				}
			}

			int pixWidth = 0;
			int pixHeight = 0;
			for (int i = tlSelectedCell.x; i <= brSelectedCell.x; i++)
				pixHeight += getCell(i, tlSelectedCell.y).getHeight();
			for (int j = tlSelectedCell.y; j <= brSelectedCell.y; j++)
				pixWidth += getCell(tlSelectedCell.x, j).getWidth();

			int height = brSelectedCell.x - tlSelectedCell.x + 1;
			int width = brSelectedCell.y - tlSelectedCell.y + 1;

			CompoundEdit edit = new RedrawingCompoundEdit();
			StateEdit e = new StateEdit(cellGrid);
			StateEdit e2 = new StateEdit(firstCell);

			fusionCells(tlSelectedCell, brSelectedCell);
			firstCell.setCellSize(new Dimension(width, height));
			firstCell.setLocation(tlSelectedCell);
			firstCell.setPreferredSize(new Dimension(pixWidth, pixHeight));

			setCell(
				firstCell,
				firstCell.getLocation().x,
				firstCell.getLocation().y);

			e.end();
			e2.end();
			edit.addEdit(e);
			edit.addEdit(e2);
			edit.end();
			undoHandler.addEdit(edit);

			tlSelectedCell = firstCell.getLocation();
			brSelectedCell = firstCell.getLocation();
			redrawCellPanel();
			cellPanel.revalidate();
			repaint();
			cellPanel.paint(cellPanel.getGraphics());
		}

		private void undo() {
			try {
				undoHandler.undo();
				updateToolBar();
				repaint();
			} catch (CannotUndoException ex) {
				System.out.println("Unable to undo: " + ex);
			}
		}

		private void redo() {
			try {
				undoHandler.redo();
				updateToolBar();
				repaint();
			} catch (CannotRedoException ex) {
				System.out.println("Unable to redo: " + ex);
			}
		}

		private void border() {
			Iterator iter = subTableList.iterator();
			/*
			 * Si la zone s?lectionn?e correspond d?j? ? un encadrement, on le retire
			 */
			while (iter.hasNext()) {
				CellArray ca = (CellArray) iter.next();
				if (getCell(ca.getTLCell())
					.getLocation()
					.equals(tlSelectedCell)
					&& getCell(ca.getBRCell()).getLocation().equals(
						brSelectedCell)) {

					removeSubTable(ca);
					repaint();
					return;
				}
			}
			createSubTable(tlSelectedCell, brSelectedCell);
			repaint();
		}
	}

	/*****************************************************************************
	 * Classe interne g?rant les actions sur les boutons des marges
	 * et jouant le role de mouselistener pour le redimensionnement 
	 * des colonnes et des lignes
	 */
	private class BorderAction
		extends MouseAdapter
		implements ActionListener, MouseMotionListener {

		private JPopupMenu popup;
		private JMenuItem addRow, insertRow, delRow, addCol, insertCol, delCol;
		private BorderButton popupButton; // bouton appelant le menu popup

		public BorderAction() {
			popup = new JPopupMenu();

			insertRow = new JMenuItem("Ins?rer une ligne");
			insertRow.addActionListener(this);
			popup.add(insertRow);

			addRow = new JMenuItem("Ajouter une ligne");
			addRow.addActionListener(this);
			popup.add(addRow);

			delRow = new JMenuItem("Supprimer la ligne");
			delRow.addActionListener(this);
			popup.add(delRow);

			insertCol = new JMenuItem("Ins?rer une colonne");
			insertCol.addActionListener(this);
			popup.add(insertCol);

			addCol = new JMenuItem("Ajouter une colonne");
			addCol.addActionListener(this);
			popup.add(addCol);

			delCol = new JMenuItem("Supprimer la colonne");
			delCol.addActionListener(this);
			popup.add(delCol);
		}

		//*********** gestion des s?lection des bouttons *********************
		public void actionPerformed(ActionEvent e) {
			updateCellHeight();

			if (e.getSource() instanceof JMenuItem) {
				JMenuItem jmi = (JMenuItem) e.getSource();

				Point pb = popupButton.getLocation();

				if (jmi == addRow)
					addRow(pb.y + 1, false);
				else if (jmi == insertRow)
					addRow(pb.y, true);
				else if (jmi == delRow)
					delRow(pb.y);
				else if (jmi == addCol)
					addCol(pb.y + 1, false);
				else if (jmi == insertCol)
					addCol(pb.y, true);
				else if (jmi == delCol)
					delCol(pb.y);
			}

			if (!(e.getSource() instanceof BorderButton))
				return;

			if (redim)
				return;

			select((BorderButton) e.getSource());

			repaint();
			firstCell.grabFocus();
		}

		private void select(BorderButton b) {
			Point p = b.getLocation();

			// bouton dans le coin sup?rieur gauche
			if (p.x == 0 && p.y == -1) {
				firstCell = getCell(0, 0);
				tlSelectedCell = new Point(0, 0);
				brSelectedCell =
					new Point(getTabSize().height - 1, getTabSize().width - 1);

			} // boutons de la barre verticale
			else if (p.x == 0) {
				firstCell = getCell(p.y, 0);

				Cell cell = getCell(p.y, getTabSize().width - 1);
				Point xy1 = firstCell.getLocation();
				Dimension hw1 = firstCell.getCellSize();
				Point xy2 = cell.getLocation();
				Dimension hw2 = cell.getCellSize();

				// en x :
				int top;
				int bottom;
				top = (xy1.x <= xy2.x ? xy1.x : xy2.x);
				bottom =
					(xy1.x + hw1.height >= xy2.x + hw2.height
						? xy1.x + hw1.height - 1
						: xy2.x + hw2.height - 1);
				// en y :
				int left = 0;
				int right = getTabSize().width - 1;

				/* algorithme it?ratif permettant de ne pas couper des cellules fusionn?es
				 * lors de la s?lection et donc d'agrandir malgr? l'utilisateur la 
				 * zone s?lectionn?e pour inclure l'enti?ret? des cellules.
				 */
				boolean isModified = true;
				Cell c;
				while (isModified) {
					isModified = false;
					for (int i = top; i <= bottom; i++) {
						for (int j = left; j <= right; j++) {
							c = getCell(i, j);
							// en x
							if (c.getLocation().x < top) {
								top = c.getLocation().x;
								isModified = true;
							}
							if (c.getLocation().x + c.getCellSize().height - 1
								> bottom) {
								bottom =
									c.getLocation().x
										+ c.getCellSize().height
										- 1;
								isModified = true;
							}
						}
					}
				}

				tlSelectedCell.setLocation(top, left);
				brSelectedCell.setLocation(bottom, right);

			} // boutons de la barre sup?rieure
			else if (p.x == 1) {
				firstCell = getCell(0, p.y);
				tlSelectedCell = new Point(0, p.y);
				brSelectedCell = new Point(getTabSize().height - 1, p.y);

				Cell cell = getCell(getTabSize().height - 1, p.y);

				Point xy1 = firstCell.getLocation();
				Dimension hw1 = firstCell.getCellSize();
				Point xy2 = cell.getLocation();
				Dimension hw2 = cell.getCellSize();

				// en x :
				int top = 0;
				int bottom = getTabSize().height - 1;

				// en y :
				int left = (xy1.y <= xy2.y ? xy1.y : xy2.y);
				int right =
					(xy1.y + hw1.width >= xy2.y + hw2.width
						? xy1.y + hw1.width - 1
						: xy2.y + hw2.width - 1);

				/* algorithme it?ratif permettant de ne pas couper des cellules fusionn?es
				 * lors de la s?lection et donc d'agrandir malgr? l'utilisateur la 
				 * zone s?lectionn?e pour inclure l'enti?ret? des cellules.
				 */
				boolean isModified = true;
				Cell c;
				while (isModified) {
					isModified = false;
					for (int i = top; i <= bottom; i++) {
						for (int j = left; j <= right; j++) {
							c = getCell(i, j);
							// en y
							if (c.getLocation().y < left) {
								left = c.getLocation().y;
								isModified = true;
							}
							if (c.getLocation().y + c.getCellSize().width - 1
								> right) {
								right =
									c.getLocation().y
										+ c.getCellSize().width
										- 1;
								isModified = true;
							}
						}
					}
				}
				tlSelectedCell.setLocation(top, left);
				brSelectedCell.setLocation(bottom, right);
			}
		}

		//**************** gestion du redimensionnement *********************
		// pour la gestion du undo-redo
		StateEdit edit;
		/* true qd l'utilisateur est pr?s du bord d'un bouton, donc on passe en
		 * mode redimensionnement possible et il y a changement de curseur.
		 */
		private boolean redim = false;
		// true qd l'utilisateur est en train de redimensionner.
		boolean resizing = false;

		public void mouseMoved(MouseEvent e) {
			BorderButton butt = (BorderButton) e.getSource();
			Point p = butt.getLocation();
			if (p.x == 0) {
				if (butt.getHeight() - e.getY() < 5) {
					setCursor(new Cursor(Cursor.N_RESIZE_CURSOR));
					redim = true;
					return;
				}
			} else {
				if (butt.getWidth() - e.getX() < 5) {
					setCursor(new Cursor(Cursor.E_RESIZE_CURSOR));
					redim = true;
					return;
				}
			}
			if (resizing)
				return;
			setCursor(Cursor.getDefaultCursor());
			redim = false;
		}

		public void mouseDragged(MouseEvent e) {
			if (resizing == false)
				return;
			BorderButton butt = (BorderButton) e.getSource();
			Point p = butt.getLocation();
			if (p.x == 0) {
				setRowSize(p.y, e.getY() - butt.getSize().height);
				//pas de updateHeight n?cessaire ici.
			} else {
				setColSize(p.y, e.getX() - butt.getSize().width);
				updateCellHeight();
			}
			cellPanel.revalidate();
			repaint();
			firstCell.grabFocus();
		}
		public void mouseReleased(MouseEvent e) {
			setCursor(Cursor.getDefaultCursor());
			redim = false;

			if (!resizing) {
				select((BorderButton) e.getSource());

				if (e.isPopupTrigger()) {
					popupButton = (BorderButton) e.getSource();
					Point p = popupButton.getLocation();
					if (p.x == 0 && p.y != -1) {
						insertRow.setVisible(true);
						addRow.setVisible(false);
						delRow.setVisible(true);
						insertCol.setVisible(false);
						addCol.setVisible(false);
						delCol.setVisible(false);
						if (p.y == getTabSize().height - 1)
							addRow.setVisible(true);
					} else if (p.x == 1 && p.y != -1) {
						insertRow.setVisible(false);
						addRow.setVisible(false);
						delRow.setVisible(false);
						insertCol.setVisible(true);
						addCol.setVisible(false);
						delCol.setVisible(true);
						if (p.y == getTabSize().width - 1)
							addCol.setVisible(true);
					}
					popup.show(e.getComponent(), e.getX(), e.getY());
					return;
				}
			}
			if (resizing == false)
				return;
			resizing = false;
			edit.end();
			undoHandler.addEdit(edit);
			repaint();
			firstCell.grabFocus();
		}
		public void mousePressed(MouseEvent e) {
			if (redim)
				resizing = true;
			edit =
				new StateEdit(
					new RedimEdit(
						((BorderButton) e.getSource()).getLocation()));
		}
		public void mouseExited(MouseEvent e) {
			if (resizing)
				return;
			setCursor(Cursor.getDefaultCursor());
			repaint();
		}
	}

	/*****************************************************************************
	 * Gestion de l'affichage des options dans la combobox du choix de la couleur
	 * de texte.
	 */
	private class ColoredText extends Component {
		Color color;
		public ColoredText(Color c) {
			color = c;
		}

		public void paint(Graphics g) {
			g.setColor(color);
			g.setFont(new Font(firstCell.getFont().getName(), Font.PLAIN, 10));
			g.drawString("Aa", 0, 10);
			g.drawLine(0, 11, 12, 11);
		}
		public Dimension getPreferredSize() {
			return new Dimension(12, 12);
		}
	}

	/*****************************************************************************
	 * Gestion de l'affichage des options dans la combobox du choix de la couleur
	 * de fond des cases.
	 */
	private class ColoredCell extends Component {
		Color color;
		public ColoredCell(Color c) {
			color = c;
		}

		public void paint(Graphics g) {
			g.setColor(color);
			g.fillRect(3, 3, 10, 10);
		}
		public Dimension getPreferredSize() {
			return new Dimension(12, 12);
		}
	}

	/*****************************************************************************
	 * UndoManager qui permet aussi de maintenir ? jour les boutons bUndo et 
	 * bRedo en fonction de canUndo() et canRedo(). Il comprend aussi une m?thode
	 * setActive() qui pemet de d?sactiver momentan?ment l'enregistrement des
	 * UndoableEdit par la m?thode addEdit(...).
	 */
	private class UndoHandler extends UndoManager {
		private boolean isActive = true;
		private boolean blocked = false;

		public void blockActiver(boolean b) {
			blocked = b;
			isActive = !b;
			update();
		}

		public void setActive(boolean b) {
			if (blocked)
				return;
			isActive = b;
			update();
		}

		public void update() {
			if (super.canUndo())
				bUndo.setEnabled(true);
			else
				bUndo.setEnabled(false);
			if (super.canRedo())
				bRedo.setEnabled(true);
			else
				bRedo.setEnabled(false);
		}

		public boolean addEdit(UndoableEdit e) {
			if (!isActive)
				return false;
			//System.out.println("addEdit");
			boolean b = super.addEdit(e);
			update();
			return b;
		}
		public void undo() {
			blockActiver(true);
			super.undo();
			blockActiver(false);
		}
		public void redo() {
			blockActiver(true);
			super.redo();
			blockActiver(false);
		}
	}

	/*****************************************************************************
	 * Classe interne g?rant le undo-redo du redimensionnement de la taille des 
	 * cellules.
	 */
	private class RedimEdit implements StateEditable {
		private Point point;

		// p: position d'un BOUTON!! (attention ? la num?rotation)
		public RedimEdit(Point p) {
			point = p;
		}

		public void storeState(Hashtable state) {
			// on modifie la hauteur d'une ligne
			if (point.x == 0) {
				Cell c;
				for (int j = 0; j < getTabSize().width; j++) {
					c = getCell(point.y, j);
					int h = (int) Math.round(c.getSize().getHeight());
					state.put(
						"cellheight " + point.y + "," + j,
						new Integer(h));
				}
				// on modifie la largeur d'une colonne
			} else {
				Cell c;
				for (int i = 0; i < getTabSize().height; i++) {
					c = getCell(i, point.y);
					int w = (int) Math.round(c.getSize().getWidth());
					state.put("cellwidth " + i + "," + point.y, new Integer(w));
				}
			}
		}

		public void restoreState(Hashtable state) {
			// on modifie la hauteur d'une ligne
			try {
				if (point.x == 0) {
					Cell c;
					for (int j = 0; j < getTabSize().width; j++) {
						c = getCell(point.y, j);
						int h =
							((Integer) state
								.get("cellheight " + point.y + "," + j))
								.intValue();
						int w = (int) Math.round(c.getSize().getWidth());
						c.setPreferredSize(new Dimension(w, h));
					}
					// on modifie la largeur d'une colonne
				} else {
					Cell c;
					for (int i = 0; i < getTabSize().height; i++) {
						c = getCell(i, point.y);
						int w =
							((Integer) state
								.get("cellwidth " + i + "," + point.y))
								.intValue();
						int h = (int) Math.round(c.getSize().getHeight());
						c.setPreferredSize(new Dimension(w, h));
					}
				}
			} catch (Exception e) {
			} //au cas improbable o? la taille n'a pas ?t? modifi?e
			cellPanel.revalidate();
		}
	}

	/*****************************************************************************
	 * Classe interne g?rant le undo-redo pour les sous-tableaux. 
	 * Elle m?morise une copie du vecteur subTableList.
	 */
	private class SubTableStateEdit implements StateEditable {

		public void storeState(Hashtable state) {
			state.put("subtablelist", subTableList.clone());
		}

		public void restoreState(Hashtable state) {
			if (state.containsKey("subtablelist")) {
				/* on clone car c'est un objet et sinon une modif ds SpreadSheet
				 * entraine une modif dans la hashtable*/
				subTableList =
					(ArrayList) ((ArrayList) state.get("subtablelist")).clone();
			}
			repaint();
		}
	}

	/*****************************************************************************
	 * Classe interne g?rant le undo-redo pour leredimensionnement dynamique du 
	 * tableau. Elle m?morise la taille du tableau et toutes les cases, plus la 
	 * position de cases. Elle m?morise ?galement le vecteur contenant les cadres
	 * et les positions de ceux-ci.
	 *
	private class RedimArrayStateEdit implements StateEditable {
		public void storeState(Hashtable state) {
			// taille du tableau
			Dimension dim = getTabSize();
			//enregistrement complet des cellule et de leur position
			for (int i = 0; i < getTabSize().height; i++) {
				for (int j = 0; j < getTabSize().width; j++) {
					Cell c = getCell(i, j);
					state.put("cell " + i + "," + j, c);
					state.put("celllocation " + i + "," + j, c.getLocation());
				}
			}
			// enregistrement des sous-tableaux et de leur taille
			state.put("subtablelist", subTableList.clone());
			
			Iterator iter = subTableList.iterator();
			while (iter.hasNext()) {
				CellArray ca = (CellArray) iter.next();
				Point tl = ca.getTLCell();
				Point br = ca.getBRCell();
				state.put("ca );
			}
		}
	
		public void restoreState(Hashtable state) {
			for (int i = tlCell.x; i <= brCell.x; i++) {
				for (int j = tlCell.y; j <= brCell.y; j++) {
					if (state.containsKey("cell " + i + "," + j)) {
						Cell c = (Cell) (state.get("cell " + i + "," + j));
						setCell(c, i, j);
	
					}
					if (state.containsKey("cellparam " + i + "," + j)) {
						Cell c = getCell(i, j);
						int[] param =
							(int[]) state.get("cellparam " + i + "," + j);
						c.setCellSize(new Point(param[0], param[1]));
						c.setLocation(new Point(param[2], param[3]));
						c.setPreferredSize(new Dimension(param[4], param[5]));
					}
				}
			}
			redrawCellPanel();
		}
	}*/

	/*****************************************************************************
	 * Classe implementant le cellPanel, il s'agit d'un JPanel auquel on a rajout?
	 * le tracage de la zone s?lection?e par l'utilisateur et le tracage de tous
	 * les sous-tableaux.
	 */
	private class CellPanel extends JPanel {
		public void paint(Graphics g) {
			super.paint(g);
			Graphics2D g2 = (Graphics2D) g;
			// tracage de la s?lection de l'utilisateur
			g2.setStroke(new BasicStroke(4));
			g2.setColor(new Color(150, 150, 220));
			// n?cessaire pour g?rer lec cellules fusionn?es
			Point tl =
				getCell(tlSelectedCell.x, tlSelectedCell.y).getLocation();
			Point br =
				getCell(brSelectedCell.x, brSelectedCell.y).getLocation();
			int x = getCell(tl).getX();
			int y = getCell(tl).getY();
			int width = getCell(br).getX() + getCell(br).getWidth() - x;
			int height = getCell(br).getY() + getCell(br).getHeight() - y;
			g2.drawRect(x, y, width, height);
			// dessin des bordures (sous-tableaux)
			g2.setStroke(new BasicStroke(2));
			g2.setColor(Color.black);
			Iterator iter = subTableList.iterator();
			while (iter.hasNext()) {
				CellArray ca = (CellArray) iter.next();
				tl = getCell(ca.getTLCell()).getLocation();
				br = getCell(ca.getBRCell()).getLocation();
				x = getCell(tl).getX();
				y = getCell(tl).getY();
				width = getCell(br).getX() + getCell(br).getWidth() - x;
				height = getCell(br).getY() + getCell(br).getHeight() - y;
				g2.drawRect(x, y, width, height);
			}
		}
	}
	/*****************************************************************************
	 * Classe interne g?rant les actions sur les touches du clavier. Elle permet
	 * d'utiliser Ctrl+z pour faire un undo et Ctrl+y pour redo. Et d'utiliser
	 * les touches directionnelles, "enter", "tab", Shift+"tab" pour se d?placer
	 * dans le tableau.
	 */
	private class KeyAction extends KeyAdapter {
		public void keyPressed(KeyEvent ke) {
			updateCellHeight();

			if (ke.getKeyChar() == 26 && undoHandler.canUndo()) {
				undoHandler.undo();
				updateToolBar();
				repaint();
				return;
			} else if (ke.getKeyChar() == 25 && undoHandler.canRedo()) {
				undoHandler.redo();
				updateToolBar();
				repaint();
				return;
			}

			Point p = firstCell.getLocation();
			Dimension s = firstCell.getCellSize();
			int code = ke.getKeyCode();

			if (code == KeyEvent.VK_ENTER || code == KeyEvent.VK_DOWN) {
				if (p.x < getTabSize().height - s.height)
					p.translate(s.height, 0);
			} else if (code == KeyEvent.VK_UP) {
				if (p.x > 0)
					p.translate(-1, 0);
			} else if (code == KeyEvent.VK_TAB) {
				if (ke.isShiftDown()) {
					if (p.y > 0)
						p.translate(0, -1);
				} else if (p.y < getTabSize().width - s.width)
					p.translate(0, s.width);
			} else if (code == KeyEvent.VK_RIGHT) {
				if (p.y < getTabSize().width - s.width)
					p.translate(0, s.width);
			} else if (code == KeyEvent.VK_LEFT) {
				if (p.y > 0)
					p.translate(0, -1);
			} else
				return;
			
			ke.consume();
			p = getCell(p).getLocation(); // ? cause des fusions
			firstCell = getCell(p);
			tlSelectedCell.setLocation(p);
			brSelectedCell.setLocation(p);
			firstCell.grabFocus();
			repaint();
			updateToolBar();
		}
	}

	public boolean getVersion() {
		return bonneVersion;
	}
	/**
	 * Classe d?finissant la repr?sentation du tableau,
	 * permettant la modification dynamique de taille.
	 */
	private class CellGrid implements StateEditable {
		private Cell[][] grid;
		private Dimension dim;
		public CellGrid(int x, int y) {
			grid = new Cell[x][y];
			dim = new Dimension(y, x); // d'abords width puis height
		}

		public Dimension getSize() {
			return new Dimension(dim);
		}

		public Cell getCell(int i, int j) {
			if (i >= dim.height || j >= dim.width)
				throw new IndexOutOfBoundsException("(" + i + "," + j + ")");

			return grid[i][j];
		}

		public void setCell(int i, int j, Cell c) {
			if (i < dim.height && j < dim.width)
				grid[i][j] = c;
			else
				throw new IndexOutOfBoundsException();
		} //x: position de la ligne ? rajouter
		public void insertRow(int x) { //si on rajoute au bout et qu'il reste de la place, on n'agrandis pas
			if (x == dim.height && grid.length > dim.height) {
				dim.height++;
				/*System.out.println(
					"first algo: dim.height: "
						+ dim.height
						+ " grid.length "
						+ grid.length);*/
				return;
			} // si on est ? la taille limite, on rajoute deux lignes
			int n = 1;
			if (dim.height == grid.length)
				n = 2;
			Cell[][] temp = new Cell[dim.height + n][dim.width];
			// copie de la partie sup?rieure du tableau:
			for (int i = 0; i < x; i++)
				System.arraycopy(grid[i], 0, temp[i], 0, dim.width);
			// copie de la partie inf?rieure
			for (int i = x; i < dim.height; i++)
				System.arraycopy(grid[i], 0, temp[i + 1], 0, dim.width);
			dim.height++;
			grid = temp;
			/*System.out.println(
			/	"dim.height: " + dim.height + " grid.length " + grid.length);*/
		}

		public void delRow(int x) { //si on enl?ve au bout, on ne change pas de taille
			if (x == dim.height - 1) {
				dim.height--;
				for (int j = 0; j < dim.width; j++)
					grid[x][j] = null;
				/*System.out.println(
					"first algo: dim.height: "
						+ dim.height
						+ " grid.length "
						+ grid.length);*/
				return;
			} // sinon, il faut refaire le tableau (on ne diminue sa taille que
			// s'il est vraiment plus grand)
			int n = grid.length;
			if (dim.height < grid.length - 4)
				n = dim.height + 1;
			Cell[][] temp = new Cell[n][dim.width];
			// copie de la partie sup?rieure du tableau:
			for (int i = 0; i < x; i++)
				System.arraycopy(grid[i], 0, temp[i], 0, dim.width);
			// copie de la partie inf?rieure
			for (int i = x + 1; i < dim.height; i++)
				System.arraycopy(grid[i], 0, temp[i - 1], 0, dim.width);
			dim.height--;
			grid = temp;
			/*System.out.println(
				"dim.height: " + dim.height + " grid.length " + grid.length);*/
		}

		public void insertCol(int y) { //si on rajoute au bout et qu'il reste de la place, on n'agrandis pas
			if (y == dim.width && grid[0].length > dim.width) {
				dim.width++;
				/*System.out.println(
					"first algo: dim.width: "
						+ dim.width
						+ " grid[0].length "
						+ grid[0].length);*/
				return;
			} // si on est ? la taille limite, on rajoute deux colonnes
			int n = 1;
			if (dim.width == grid[0].length)
				n = 2;
			Cell[][] temp = new Cell[dim.height][dim.width + n];
			for (int i = 0; i < dim.height; i++) {
				// copie de la partie gauche du tableau:
				System.arraycopy(grid[i], 0, temp[i], 0, y);
				// copie de la partie droite du tableau:
				System.arraycopy(grid[i], y, temp[i], y + 1, dim.width - y);
			}
			dim.width++;
			grid = temp;
			/*System.out.println(
				"dim.width: "
					+ dim.width
					+ " grid[0].length "
					+ grid[0].length);*/
		}

		public void delCol(int y) { //si on enl?ve au bout, on ne change pas de taille
			if (y == dim.width - 1) {
				dim.width--;
				for (int i = 0; i < dim.height; i++)
					grid[i][y] = null;
				/*System.out.println(
					"first algo: dim.width: "
						+ dim.width
						+ " grid[0].length "
						+ grid[0].length);*/
				return;
			} // sinon, il faut refaire le tableau (on ne diminue sa taille que
			// s'il est vraiment plus grand)
			int n = grid[0].length;
			if (dim.width < grid[0].length - 4)
				n = dim.width + 1;
			Cell[][] temp = new Cell[dim.height][n];
			for (int i = 0; i < dim.height; i++) {
				// copie de la partie gauche du tableau:
				System.arraycopy(grid[i], 0, temp[i], 0, y);
				// copie de la partie droite du tableau:
				System.arraycopy(grid[i], y + 1, temp[i], y, dim.width - y - 1);
			}

			dim.width--;
			grid = temp;
			/*System.out.println(
				"dim.width: "
					+ dim.width
					+ " grid[0].length "
					+ grid[0].length);*/
		}

		public void storeState(Hashtable state) {
			state.put("dim", new Dimension(dim));
			for (int i = 0; i < dim.height; i++) {
				for (int j = 0; j < dim.width; j++) {
					state.put("cell " + i + "," + j, grid[i][j]);
				}
			}
		}

		public void restoreState(Hashtable state) {
			Cell[][] newGrid;

			/* Si la taille a chang?, on cr?e un nouveau tableau,
			 * sinon, on utilise le tableau courant. Ensuite, on met 
			 * dans le tableau les cellules si elles sont dans la hashtable
			 */
			if (state.containsKey("dim")) {
				dim = new Dimension((Dimension) state.get("dim"));
				newGrid = new Cell[dim.height][dim.width];
			} else
				newGrid = grid;

			for (int i = 0; i < dim.height; i++) {
				for (int j = 0; j < dim.width; j++) {
					if (state.containsKey("cell " + i + "," + j))
						newGrid[i][j] = (Cell) state.get("cell " + i + "," + j);
					else
						newGrid[i][j] = grid[i][j];
				}
			}
			grid = newGrid;

			firstCell = getCell(0, 0);
			tlSelectedCell = new Point(0, 0);
			brSelectedCell = new Point(0, 0);
		}
	}

	/**
	 * Classe ?tendant un CompoundEdit. Celui-ci effectue un redrawing
	 * du tableau apr?s le undo et le redo.
	 */
	private class RedrawingCompoundEdit extends CompoundEdit {
		public void undo() {
			super.undo();
			redrawCellPanel();
		}
		public void redo() {
			super.redo();
			redrawCellPanel();
		}
	}
}
